import platform
import pyglet

# ---------------------------------------------------------------
# On OS X we need to disable the shadow context
# because the 2.1 shadow context cannot be upgrade to a 3.3+ core
if platform.system() == "Darwin":
    pyglet.options["shadow_window"] = False
# ---------------------------------------------------------------

import cProfile

from ..loader.loader import ResourceLoader
from ..gpu.opengl_data import OpenGLData
from ..systems.scene_system import SceneSystem


# ========================================================
# 21/06/2021 :
# - dynamic sprites = 3k
# - static  sprites = 100k
# both calling the sprite.update() method to check if the
# array.array must be copied into the GPU texture
# ========================================================


# ========================================================
# DEBUG PARAMS
# ========================================================
DEBUG_DISPLAY_PERFS  = False


class Pyrogen():

    def __init__(self,
                 atlasDir=".",
                 fsGpuMemSize= 8*1024*1024*4,
                 *args, **kwargs):

        # Create GL config
        cfg = pyglet.gl.Config(
            major_version=3,
            minor_version=3,
            forward_compatible=True,
            depth_size=24,
            double_buffer=True,
            sample_buffers=1,
            samples=4,
        )

        # Instanciate Window
        self._pyrogenWin = PyrogenWindow(atlasDir=atlasDir,
                                     fsGpuMemSize=fsGpuMemSize,
                                     config=cfg,
                                     *args, **kwargs)

    def addImage(self, name, path, spriteInfo=None):
        self._pyrogenWin.addImage(name, path, spriteInfo)

    def addFont(self, name, path, size=128, border=1):
        self._pyrogenWin.addFont(name, path, size, border)

    def finalizeResources(self):
        self._pyrogenWin.finalizeResources()

    def addScene(self, name, ref):
        self._pyrogenWin.addScene(name, ref)

    def run(self):
        self._pyrogenWin.run()




class PyrogenWindow(pyglet.window.Window):

    # ========================================================
    #  EVENTS
    # ========================================================
    def _keyboardEvent(self, keyID, isPressed, modifiers):
        #print(f"<KEY> id={keyID} isPressed={isPressed} modifiers={modifiers}")
        self._scnMgr.keyboardEvent(keyID, isPressed, modifiers)
    def _mouseButtonEvent(self, x, y, buttonID, isPressed, modifiers):
        y = self.height-y
        #print(f"<MOUSE-BUTTON> position=({x},{y}) buttonID={buttonID} isPressed={isPressed} modifiers={modifiers}")
        self._scnMgr.mouseButtonEvent(x, y, buttonID, isPressed, modifiers)
    def _mouseMotionEvent(self, x, y, dx, dy):
        dy = -dy
        #print(f"<MOUSE-MOVE> position=({x},{y}) direction=({dx},{dy})")
        self._scnMgr.mouseMotionEvent(x, y, dx, dy)
    def _mouseDragEvent(self, x, y, dx, dy, buttonID, modifiers):
        dy = -dy
        #print(f"<MOUSE-DRAG> position=({x},{y}) direction=({dx},{dy}) buttonID={buttonID} modifiers={modifiers}")
        self._scnMgr.mouseDragEvent(x, y, dx, dy, buttonID, modifiers)
    def _mouseScrollEvent(self, x, y, dx, dy):
        #print(f"<MOUSE-SCROLL> position=({x},{y}) direction=({dx},{dy})")
        self._scnMgr.mouseScrollEvent(x, y, dx, dy)
    def _gamepadButtonEvent(self, gamepadID, buttonID, isPressed):
        #print(f"<GAMEPAD-BUTTON> gamePadID=({gamepadID}) buttonID=({buttonID}) isPressed={isPressed}")
        self._scnMgr.gamepadButtonEvent(gamepadID, buttonID, isPressed)
    def _gamepadAxisEvent(self, gamepadID, axisID, analogValue):
        if axisID == "z":
            analogValue *= -1
        #print(f"<GAMEPAD-AXIS> gamePadID=({gamepadID}) axisID=({axisID}) value={analogValue}")
        self._scnMgr.gamepadAxisEvent(gamepadID, axisID, analogValue)

    # ========================================================
    #  PYGLET CALLBACKS
    # ========================================================
    def on_key_press(self, symbol, modifiers):
        self._keyboardEvent(symbol, True, modifiers)
        if symbol == pyglet.window.key.ESCAPE:
            self._openGlData.displayFS()
            self.close()
    def on_key_release(self, symbol, modifiers):
        self._keyboardEvent(symbol, False, modifiers)
    def on_mouse_press(self, x, y, button, modifiers):
        self._mouseButtonEvent(x, y, button, True, modifiers)
    def on_mouse_release(self, x, y, button, modifiers):
        self._mouseButtonEvent(x, y, button, False, modifiers)
    def on_mouse_motion(self, x, y, dx, dy):
        self._mouseMotionEvent(x, y, dx, dy)
    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        self._mouseDragEvent(x, y, dx, dy, buttons, modifiers)
    def on_mouse_scroll(self, x, y, scroll_x, scroll_y):
        self._mouseScrollEvent(x, y, scroll_x, scroll_y)
    def on_joystick_press(self, joystick, button):
        joyID = self.gamepads[joystick]
        self._gamepadButtonEvent(joyID, button, True)
    def on_joystick_release(self, joystick, button):
        joyID = self.gamepads[joystick]
        self._gamepadButtonEvent(joyID, button, False)
    def on_joystick_motion(self, joystick, axis, value):
        joyID = self.gamepads[joystick]
        self._gamepadAxisEvent(joyID, axis, value)
    def on_joystick_hat(self, joystick, x, y):
        joyID = self.gamepads[joystick]
        self._gamepadAxisEvent(joyID, "x",  x)
        self._gamepadAxisEvent(joyID, "y", -y)

    # ========================================================
    #  GAMEPADS
    # ========================================================
    def _loadGamepads(self):
        # Gamepad callbacks
        joysticks = pyglet.input.get_joysticks()
        N = len(joysticks)
        s1 = "is" if N <= 1 else "are"
        s2 = "" if N <= 1 else "s"
        if N == 0:
            N = "no"
        print(f"\nThere {s1} {N} connected gamepad{s2}")
        self.gamepads = {}
        if len(joysticks) > 0:
            i = 0
            for g in joysticks:
                g.open()
                if g not in self.gamepads:
                    self.gamepads[g] = i
                    i += 1
                g.on_joybutton_press   = self.on_joystick_press
                g.on_joybutton_release = self.on_joystick_release
                g.on_joyaxis_motion    = self.on_joystick_motion
                g.on_joyhat_motion     = self.on_joystick_hat
                print(f"    ID : #{i-1} - NAME : {g.device.name}")

    # ========================================================
    #  PERFORMANCE DISPLAY (DEBUG)
    # ========================================================
    def _displayPerfs(self, cpr):
        # List methods under supervision
        watchMethods = [
            "updateScript",
            #"render",
            #"update",
            #"updateMovingSprites",
            #"write",
            #"_write",
            #"_verifCHK",
            #"_computeCHK",
            #"_explodeID",
            #"genVertex",
            #"writeBlock",
            #"getData",
            #"on_draw"
        ]
        watchFiles = [
            "script_system",
            "input_system",
            "gfx_system",
            #"gfx_components",
        ]
        # Get stats of execution : parse them
        st = cpr.getstats()
        out = []
        for row in st:
            # Prepare data
            code = str(row[0])
            code = code.replace("code object ", "")
            code = code.replace("built-in ", "")
            code = code.replace("method ", "")
            code = code.replace(" objects", "")
            code = code.replace("<", "")
            code = code.replace(">", "")
            code = code.split(",")
            method = code[0].split(" ")[0]
            file = code[1] if len(code) > 1 else ""
            file = file.replace("file", "")
            file = file.replace("\"", "")
            #            file   = os.path.basename(file).split(".")[0]
            line = code[2] if len(code) > 2 else ""
            line = line.lower().replace("line ", "")
            # store data
            toBeStored = False
            for w in method.split(" "):
                if w in watchMethods:
                    toBeStored = True
                    break
            for w in watchFiles:
                if w in file:
                    toBeStored = True
            #            toBeStored = True
            if toBeStored:
                out.append({"method": method,
                            "file": file,
                            "line": line,
                            "ncalls": str(row[1]).replace(".", ","),
                            "tottime": str(1000 * row[4]).replace(".", ","),
                            "totpercall": str(1000 * row[4] / row[1]).replace(".", ","),
                            "cumtime": str(1000 * row[3]).replace(".", ","),
                            "cumpercall": str(1000 * row[3] / row[1]).replace(".", ","),
                            })
        out = sorted(out, key=lambda x: (x["file"], x["method"]))
        print("PERFS")
        for o in out:
            me = o["method"]
            fi = o["file"]
            li = o["line"]
            nc = o["ncalls"]
            tt = o["tottime"]
            tp = o["totpercall"]
            ct = o["cumtime"]
            cp = o["cumpercall"]
            print(f"{fi}/{me} ({li})\t{nc}\t{tt}\t{tp}\t{ct}\t{cp}")

    # ========================================================
    #  CONSTRUCTOR
    # ========================================================
    def __init__(self,
                 atlasDir=".",
                 fsGpuMemSize= 8*1024*1024*4,
                 *args, **kwargs, ):
        # Parent constructor
        super().__init__(*args, **kwargs)
        # Gamepads
        self._loadGamepads()
        # initiate the resource loader
        self._loader = ResourceLoader(atlasDir)
        # GL Stuff
        self._openGlData = OpenGLData(self._loader, fsGpuMemSize)
        # Profiling
        self._FPS = []
        # Create scene manager
        self._scnMgr = SceneSystem()

    # ========================================================================
    #  RESOURCES
    # ========================================================================
    def addImage(self, name, path, spriteInfo=None):
        self._loader.addImage(name, path, spriteInfo)
    def addFont(self, name, path, size=128, border=1):
        self._loader.addFont(name, path, size=size, border=border)
    def finalizeResources(self):
        self._openGlData.createAtlas()
    def addScene(self, name, ref):
        self._scnMgr.registerScene(name, ref)

    # ========================================================
    #  UPDATE METHOD
    # ========================================================
    def update(self, deltaTime):
        # Process FPS
        self._FPS.append(deltaTime)
        if len(self._FPS)==60:
            print(f">>>>>>>>>>>>> FPS = {round(60/sum(self._FPS),2)} <<<<<<<<<<<<<<<<<<<<<<<")
            self._FPS = []

        # Update the current scene
        # this will update the world and all the systems inside
        self._scnMgr.updateScene(deltaTime)

    # ========================================================
    #  RENDER METHOD
    # ========================================================
    def on_draw(self):
        # Render the current scene
        self._scnMgr.renderScene()

    # ========================================================
    #  RESIZE APPLICATION WINDOW
    # ========================================================
    def on_resize(self, width, height):
        pass
#        self.ctx.screen.viewport = 0, 0, *self.get_framebuffer_size()
#  check window.get_pixel_ratio() to be sure of the physical screen size


    # ========================================================
    #  MAIN LOOP
    # ========================================================
    def run(self):
        # Do not run the app if no scene has been created
        if self._scnMgr.getNbScenes() <= 0:
            raise RuntimeError("[ERROR] Impossible to launch the application because there is no Scene registered !")

        # update loop interval
        pyglet.clock.schedule_interval(self.update, 1/65)

        # Start pyglet app and profile it if needed
        cpr = None
        if DEBUG_DISPLAY_PERFS:
            cpr = cProfile.Profile()
            cpr.enable()
        pyglet.app.run()
        if DEBUG_DISPLAY_PERFS:
            cpr.disable()
            self._displayPerfs(cpr)




import array
import sys


class FsGpuBuffer():

    CHECK_INTEGRITY = False

    __slots__ = ['_buffer',
                 '_size',
                 '_nbComp',
                 '_modified',
                 '_freeBlocks',
                ]

    # ----------------------------------------------------
    # CONSTANTS
    # ----------------------------------------------------
    FREE = 0

    LENG = 0        # length of data block
    TYPE = 1        # Type of block (empty, sprite, rectangle, oval, line, text, light, ...)
                    # Types are not handled internally but provided from outside (except the 0/empty value)
    SIZE = 2        # user data size (can be less or equal to LENG)
    CHCK = 3        # Block Checksum

    OVERHEAD = 4    # 4 values as overhead
    MIN_SIZE = 4    # At least 4 values for a user block (that means 50 useful data)


    # ----------------------------------------------------
    # INTEGRITY CHECK
    # ----------------------------------------------------
    def _verifOffset(self, offset):
        if FsGpuBuffer.CHECK_INTEGRITY and offset > self._size:
            dump = self.dump()
            raise RuntimeError(f"[ERROR] bad block offsey @{offset} \n{dump}")

    def _computeCHK(self, offset):
        v = 0
        if FsGpuBuffer.CHECK_INTEGRITY:
            v += self._buffer[offset + FsGpuBuffer.LENG]
            v += self._buffer[offset + FsGpuBuffer.TYPE]
            v += self._buffer[offset + FsGpuBuffer.SIZE]
        return v

    def _verifCHK(self, offset):
        if FsGpuBuffer.CHECK_INTEGRITY:
            # Check offset
            self._verifOffset(offset)
            # compute checksum
            v = self._computeCHK(offset)
            if self._buffer[offset + FsGpuBuffer.CHCK] != v:
                dump = self.dump()
                raise RuntimeError(f"[ERROR] bad data block integrity (checksum) @offset={offset} \n{dump}")
            if self._buffer[offset + FsGpuBuffer.SIZE] > self._buffer[offset + FsGpuBuffer.LENG]:
                dump = self.dump()
                raise RuntimeError(f"[ERROR] bad data block integrity (lengths) @offset={offset} \n{dump}")


    # ----------------------------------------------------
    # FREE BLOCKS
    # ----------------------------------------------------
    # This must be done when AFTER the block has been freed
    def _addFreeBlock(self, offset):
        # Check offset
        self._verifCHK(offset)
        # Get type
        T = self._buffer[offset + FsGpuBuffer.TYPE]
        if not T == FsGpuBuffer.FREE:
            raise RuntimeError(f"[ERROR] impossible to add a free block as it is not free ! offset={offset}")
        # Get user Length
        L = self._buffer[offset + FsGpuBuffer.LENG]
        # check entry
        if offset in self._freeBlocks:
            raise RuntimeError(f"[ERROR] impossible to add a free block as it is already added ! offset={offset}")
        # add entry
        self._freeBlocks[offset] = L

    # This must be done BEFORE filling it or AFTER erasing it
    def _removeFreeBlock(self, offset):
        # Check offset
        self._verifCHK(offset)
        # Get type
        T = self._buffer[offset + FsGpuBuffer.TYPE]
        if not T == FsGpuBuffer.FREE:
            dmp = self.dump()
            raise RuntimeError(f"[ERROR] impossible to remove a free block as it is not free yet ! offset={offset} T={T}\n"+dmp)
        # check entry
        if offset not in self._freeBlocks:
            raise RuntimeError(f"[ERROR] impossible to remove a free block as it is not added ! offset={offset}")
        # remove entry
        del self._freeBlocks[offset]

    # update is done AFTER the block has been updated
    def _updateFreeBlock(self, offset, newLength):
        # Check offset
        self._verifCHK(offset)
        # Get type
        T = self._buffer[offset + FsGpuBuffer.TYPE]
        if not T == FsGpuBuffer.FREE:
            raise RuntimeError(f"[ERROR] impossible to update a free block as it is not free yet ! offset={offset}")
        # check entry
        if offset not in self._freeBlocks:
            raise RuntimeError(f"[ERROR] impossible to update a free block as it is not added ! offset={offset}")
        # updateentry
        self._freeBlocks[offset] = newLength


    # ----------------------------------------------------
    # CONSTRUCTOR
    # ----------------------------------------------------
    def __init__(self, W, H=1, nbComp=4):
        # init buffer
        self._size   = W * H * nbComp
        self._nbComp = nbComp

        # > array array
        self._buffer = array.array("f", [0.0,] * int(self._size))

        # Set first block as empty
        self._buffer[FsGpuBuffer.TYPE] = FsGpuBuffer.FREE
        self._buffer[FsGpuBuffer.LENG] = self._size
        self._buffer[FsGpuBuffer.SIZE] = self._size
        self._buffer[FsGpuBuffer.CHCK] = self._computeCHK(0)

        # Modified
        self._modified = True

        # Keep a list of unused areas
        # dict with offset as key and user size as value
        self._freeBlocks = {}
        self._addFreeBlock(0)


    # ----------------------------------------------------
    # PROPERTIES
    # ----------------------------------------------------
    def getData(self):
        # > array.array
        return self._buffer

    def getBufferSize(self):
        return self._size
    def isModified(self):
        return self._modified


    # ----------------------------------------------------
    # PRIVATE METHODS
    # ----------------------------------------------------
    def _isFree(self, offset):
        T = self._buffer[offset + FsGpuBuffer.TYPE]
        return T == FsGpuBuffer.FREE

    def _isAllocated(self, offset):
        return not self._isFree(offset)

    def _freeBlock(self, offset):
        # Check data integrity
        self._verifCHK(offset)
        # Get current slot information
        if self._isFree(offset):
            raise RuntimeError(f"[ERROR] the block offset {offset} cannot be released as it is ALREADY empty !")
        # Free the block
        self._buffer[offset + FsGpuBuffer.TYPE] = FsGpuBuffer.FREE
        # Do not update block length as it must remain unchanged, but update user size
        self._buffer[offset + FsGpuBuffer.SIZE] = self._buffer[offset + FsGpuBuffer.LENG]
        # recompute new integrity
        self._buffer[offset + FsGpuBuffer.CHCK] = self._computeCHK(offset)
        # Add this block into the free list
        self._addFreeBlock(offset)
        # Try to merge with next buffer if empty too
        self._mergeNext(offset)
        # Free process is ok
        self._modified = True
        return True

    # allocate selected block
    def _allocateSelected(self, userSize, userType, selected, minSize):
        # Set local variables
        L1        = minSize
        offset    = selected
        blockSize = userSize + FsGpuBuffer.OVERHEAD
        # Check if the remaining length is at least enough for another small block
        if L1 - blockSize < FsGpuBuffer.OVERHEAD + FsGpuBuffer.MIN_SIZE:
            # not enough room : take all the room
            blockSize = L1
        # Remove the free block from the list
        self._removeFreeBlock(offset)
        # set the data
        # First set user length and type
        self._buffer[offset + FsGpuBuffer.SIZE] = userSize
        self._buffer[offset + FsGpuBuffer.TYPE] = userType
        # Check if there is at least 4 user bytes available
        if blockSize + FsGpuBuffer.MIN_SIZE > L1:
            blockSize = L1
            self._buffer[offset + FsGpuBuffer.LENG] = blockSize
            self._buffer[offset + FsGpuBuffer.CHCK] = self._computeCHK(offset)
        else:
            self._buffer[offset + FsGpuBuffer.LENG] = blockSize
            self._buffer[offset + FsGpuBuffer.CHCK] = self._computeCHK(offset)
            # Set next block info
            offset2 = offset + blockSize
            self._buffer[offset2 + FsGpuBuffer.TYPE] = FsGpuBuffer.FREE
            self._buffer[offset2 + FsGpuBuffer.LENG] = L1 - blockSize
            self._buffer[offset2 + FsGpuBuffer.SIZE] = L1 - blockSize
            self._buffer[offset2 + FsGpuBuffer.CHCK] = self._computeCHK(offset2)
            # Add the new free block in the list
            self._addFreeBlock(offset2)
        # end of process
        self._modified = True
        return offset

    # Allocation process based on free block list : complexity depends on fragmentation
    def _allocateBlock2(self, userSize, userType):
        # Get usersize multiple of nbcomps
        userSize = ((userSize+self._nbComp-1)//self._nbComp)*self._nbComp
        # Compute blocksize
        blockSize = userSize + FsGpuBuffer.OVERHEAD
        # Browse each free block from the list
        for offset in self._freeBlocks:
            # Check data integrity
            self._verifCHK(offset)
            # Check it is really free
            if not self._isFree(offset):
                raise RuntimeError(f"[ERROR] there is an error in the free block list : the block @{offset} ith length={L} is not free !")

            # Get length of this free block
            L1 = self._freeBlocks[offset]
            L2 = int(self._buffer[offset + FsGpuBuffer.LENG])
            if L1 != L2:
                raise RuntimeError(
                    f"[ERROR] there is an error in the free block list : the block @{offset} : length is not good L1={L1} L2={L2}!")

            # First check the block size fits
            if blockSize < FsGpuBuffer.OVERHEAD + FsGpuBuffer.MIN_SIZE:
                blockSize = FsGpuBuffer.OVERHEAD + FsGpuBuffer.MIN_SIZE
            # If it finally fits, allocate
            if blockSize <= L1:
                return self._allocateSelected(userSize, userType, offset, L1)

        # If we reach this part of code, no space was available
        # either the memory is full or there are some free blocks
        # too small for the requested usersize
        return None

    def _mergeNext(self, offset):
        while True:
            # Get block data length
            L = self._buffer[offset + FsGpuBuffer.LENG]
            # Get next block offset
            offset2 = int(offset + L)
            # Check data integrity of next block
            self._verifCHK(offset2)
            # Check if this new block is empty so we can merge. else we just return
            if self._isFree(offset2):
                # Get next block length
                L2 = self._buffer[offset2 + FsGpuBuffer.LENG]
                # Clear next
                self._buffer[offset2 + FsGpuBuffer.TYPE] = 0
                self._buffer[offset2 + FsGpuBuffer.LENG] = 0
                self._buffer[offset2 + FsGpuBuffer.SIZE] = 0
                self._buffer[offset2 + FsGpuBuffer.CHCK] = 0
                # Remove next block from list
                self._removeFreeBlock(offset2)

                # Update current one (just length and integrity)
                self._buffer[offset + FsGpuBuffer.LENG] = L + L2
                self._buffer[offset + FsGpuBuffer.SIZE] = L + L2
                self._buffer[offset + FsGpuBuffer.CHCK] = self._computeCHK(offset)
                # Update length of current block
                self._updateFreeBlock(offset, L+L2)
                # print(f"merged between @{offset} and @{offset2}")
                return True
            else:
                return False

    # read data (up to one full block)
    def _read(self, offset, length=-1, subOffset=0):
        # Check integrity for this block
        self._verifCHK(offset)
        # if length == -1, that means read all the block
        S = self._buffer[offset + FsGpuBuffer.SIZE]
        if length < 0:
            length = S
        # check the requested length is correct
        if length > S - subOffset:
            raise RuntimeError(f"[WARNING] reading too much from buffer - offset={offset} - readLen={length} - subOffset={subOffset}",file=sys.stderr)
        # Read the buffer part and return
        start = int(offset + FsGpuBuffer.OVERHEAD)
        end   = int(start + length)
        out = self._buffer[start:end]
        return out

    # Write data (full block)
    def _write(self, offset, values, subOffset=0):
        length = len(values)
        # Check integrity for this block
        self._verifCHK(offset)
        # Get user length of block
        S = self._buffer[offset + FsGpuBuffer.SIZE]
        # check the requested length is correct
        if length > S - subOffset:
            raise RuntimeError(f"[WARNING] writing too much to buffer - offset={offset} - writeLen={length} - subOffset={subOffset}",file=sys.stderr)
        # prepare copy position into the buffer
        start = offset+FsGpuBuffer.OVERHEAD+subOffset
        end   = start+length
        # copy sub array.array
        self._buffer[start:end] = values
        # buffer has been modified
        self._modified = True

    # ----------------------------------------------------
    # PUBLIC API
    # ----------------------------------------------------
    # this method checks if there is a slot in the allocation table
    # and checks if there is a big enough data block.
    # It returns the block offset( in the current buffer) or None if no space
    def alloc(self, userSize, userType):
        if userType == FsGpuBuffer.FREE:
            raise RuntimeError(f"[ERROR] cannot allocate a block with a 'FREE' type. Type of block must be different from the value '{FsGpuBuffer.FREE}'")
        return self._allocateBlock2(userSize, userType)

    # This method releases a previous reserved block
    def free(self, offset):
        return self._freeBlock(offset)

    # Method used to reset the "modified" flag
    def resetModify(self):
        self._modified = False

    # read values
    def read(self, offset, length=-1, subOffset=0):
        return self._read(offset, length, subOffset)

    # write values (array.array)
    def write(self, offset, values, subOffset=0):
        self._write(offset, values, subOffset)

    # Defrag operation
    # For the moment we only merge free contiguous blocks,
    # so this is not a real defrag process yet
    def defrag(self):
        # Each time we browse the free block list, if we merge free blocks,
        # the list is modified while iterating on it : this generates an exception
        # So to avoid problems, the process could stop the loop after one merge
        # has been performed. That would mean we would only merge once per frame
        # in the worst case : this would solve both the modification/iteration issue
        # and would limit the CPU load : it seems to be a good compromise
        for offset in self._freeBlocks:
            # Check data integrity
            self._verifCHK(offset)
            # Get current slot information
            if not self._isFree(offset):
                raise RuntimeError(f"[ERROR] the block offset {offset} cannot be merged as it is not empty : why is it in the free block list !?")
            # merge was done, we exit from the defrag process
            res = self._mergeNext(offset)
            if res:
                return True
        # We reached the end of the free blocks and none was merged
        return False


    # ----------------------------------------------------
    # DEBUG
    # ----------------------------------------------------
    def display(self, displayData=False):
        print(self.dump(displayData))

    def dump(self, displayData=False, offset=0):
        CR = "\n"
        msg = ""
        i = 0
        # Dump free block list
        for frOf7 in self._freeBlocks:
            L = int(self._freeBlocks[frOf7])
            msg += f"<FreeBlock @{'0x{0:0{1}X}'.format(frOf7,8)} - L={'0x{0:0{1}X}'.format(L,8)}>" + CR
        # Dump block data
        while offset < self._size:
            T = int(self._buffer[offset + FsGpuBuffer.TYPE])
            L = int(self._buffer[offset + FsGpuBuffer.LENG])
            U = int(self._buffer[offset + FsGpuBuffer.SIZE])
            C =     self._buffer[offset + FsGpuBuffer.CHCK]
            msg += f"<Block #{i} @{'0x{0:0{1}X}'.format(offset,8)} - blockLen={'0x{0:0{1}X}'.format(L,8)} - type={T} - userLen={'0x{0:0{1}X}'.format(U,8)} - checksum={C}>" + CR
            if displayData and T != FsGpuBuffer.FREE:
                data = []
                of7 = offset + FsGpuBuffer.OVERHEAD
                for i in range(U):
                    v = self._buffer[of7+i]
                    data.append(str(v))
                msg += "    " + "\n    ".join(data) + CR
            offset += int(L)
            i += 1
        return msg




from ..components.script import Script


class GfxFollowGfx(Script):

    def __init__(self, gfxTarget, gfxFollower, offset=(0,0), setAngle=False, offAngle=0, priority=0, name="GfxFollowGfx"):
        super().__init__(priority, name)
        self._target   = gfxTarget
        self._follower = gfxFollower
        self._setAngle = setAngle
        self._offset   = offset
        self._offAngle = offAngle

    def updateScript(self, deltaTime, systemTime):
        # Take target position
        x,y = self._target.getPosition()
        # Add offset
        x += self._offset[0]
        y += self._offset[1]
        # Set follower position
        self._follower.setPosition(x, y)
        # Check if we must set the angle too
        if self._setAngle:
            # Get target angle
            angle = self._target.getAngle()
            # add angle offset
            angle += self._offAngle
            # Set follower angle
            self._follower.setAngle(angle)






class Entity():

    def __init__(self, name="Entity"):
        self._name = name
        self._scene = None
        self._componentsByName = {}
        self._componentsByRef  = []

    def setScene(self, scn):
        if self._scene != None:
            raise RuntimeError("[ERROR] Impossible to attach a scene to this entity !")
        self._scene = scn
        # Register components if some have already been added
        for c in self._componentsByRef:
            self._scene.registerComponent(c)

    def resetScene(self, scn):
        if self._scene != scn:
            raise RuntimeError("[ERROR] Impossible to remove a scene link !")
        self._scene = None
        # Unregister components if needed
        for c in self._componentsByRef:
            self._scene.unregisterComponent(c)

    def getName(self):
        return self._name

    def getScene(self):
        return self._scene

    def destroy(self):
        # Remove all components
        for c in self._componentsByRef:
            self.removeComponent(c)
        # Remove the entity from the scene (if needed)


    def addComponent(self, ref):
        if ref in self._componentsByRef:
            raise RuntimeError("[ERROR] Impossible to add a component into this entity !")
        # Add component
        self._componentsByRef.append(ref)
        self._componentsByName[ref.getName()] = ref
        # Create link between component and entity
        ref.setEntity(self)
        # Register this component if a scene has been already attached
        if self._scene != None:
            self._scene.registerComponent(ref)

    def removeComponent(self, ref):
        if ref not in self._componentsByRef:
            raise RuntimeError("[ERROR] Impossible to remove a component from this entity !")
        # Unregister from the scene (if needed)
        if self._scene != None:
            self._scene.unregisterComponent(ref)
        # Remove link between component and entity
        ref.resetEntity(self)
        # Forget Component
        self._componentsByRef.remove(ref)
        del self._componentsByName[ref.getName()]


    def getNbComponents(self):
        return len(self._componentsByRef)

    def getComponentsByName(self, name):
        out = None
        if name in self._componentsByName:
            out = self._componentsByName[name]
        return out

    def getComponentsByType(self, type):
        out = []
        for c in self._componentsByRef:
            if c.getType() == type:
                out.append(c)
        return out



import array
import sys

from PIL import Image, ImageFont, ImageDraw
from os.path import exists
from os import stat
import hashlib
import pickle
from ..components.gfx import Gfx
from ..loader.bin_packing import BinPacking, UniqueID


# TODO : for each texture, if not requested, add a dummy "normal" and dummy "specular" texture
# In fact we could use small dummy "normal" and "specular" textures and reference them
# by default for ALL the sprites


# =====================================================================
# IMAGE / SPRITE SHEET
# =====================================================================


class ResourceImage():

    def __init__(self, name, filePath, id, spriteInfo=None, fontInfo=None):
        # ID
        self._id = id
        # Store base name
        self._name = name
        # Position
        self._x = 0
        self._y = 0

        # Create image from fontInfo
        if fontInfo != None:
            c = fontInfo["character"]
            x = fontInfo["x"]
            y = fontInfo["y"]
            w = fontInfo["w"]
            h = fontInfo["h"]
            # font info contains the box properties
            imgFont = Image.open(filePath)
            img = imgFont.crop( (x,y,x+w,y+h) )
            self._img = img
        else:
            # Load Pillow image into memory and set RGBA mode
            img = Image.open(filePath)
            self._img = img.convert("RGBA")
        # SpriteSheet config (by default only 1 sprite, full Size)
        self._nbSpriteX = 1
        self._nbSpriteY = 1
        self._spriteW   = self.width
        self._spriteH   = self.height
        # Configure the sprite sheet if requested
        if spriteInfo != None:
            self._nbSpriteX = spriteInfo[0]     # nb sprite frames along X-axis
            self._nbSpriteY = spriteInfo[1]     # nb sprite frames along Y-axis
            self._spriteW   = spriteInfo[2]     # Width of each sprite frame
            self._spriteH   = spriteInfo[3]     # Height of each sprite frame
        # Compute and store hash (based on file properties and image size
        st = stat(filePath)
        h = hashlib.sha256()
        h.update(name.encode()             )     # name
        h.update(filePath.encode()         )     # file path
        h.update(bytearray(st.st_size)     )     # file size
        h.update(str(st.st_mtime).encode() )     # last modification date
        h.update(bytearray(self.width)     )     # image width
        h.update(bytearray(self.height)    )     # image height
        h.update(bytearray(self._nbSpriteX))    # nb sprite frames along X-axis
        h.update(bytearray(self._nbSpriteY))    # nb sprite frames along Y-axis
        h.update(bytearray(self._spriteW  ))    # Width of each sprite frame
        h.update(bytearray(self._spriteH  ))    # Height of each sprite frame
        d = h.hexdigest()
        # keep only the 32 middle characters
        self._hash = d[16:48]

    @property
    def x0(self):
        return self._x
    @property
    def y0(self):
        return self._y
    @x0.setter
    def x0(self,v):
        self._x = v
    @y0.setter
    def y0(self,v):
        self._y = v

    @property
    def width(self):
        return self._img.size[0]
    @property
    def height(self):
        return self._img.size[1]
    @property
    def x1(self):
        return self.x0 + self.width -1
    @property
    def y1(self):
        return self.y0 + self.height -1

    @property
    def spriteWidth(self):
        return self._spriteW
    @property
    def spriteHeight(self):
        return self._spriteH


    @property
    def data(self):
        return self._img
    @property
    def name(self):
        return  self._name
    @property
    def hash(self):
        return self._hash
    @property
    def id(self):
        return self._id

    @property
    def nbSpriteX(self):
        return self._nbSpriteX
    @property
    def nbSpriteY(self):
        return self._nbSpriteY
    @property
    def spriteW(self):
        return self._spriteW
    @property
    def spriteH(self):
        return self._spriteH

    def __str__(self):
        return f"<Image id:{self.id} name:'{self.name}' - x:{self.x0} - y:{self.y0} - w:{self.width} - h:{self.height} - nbX:{self.nbSpriteX} - nbY:{self.nbSpriteY} - hash:'{self.hash}'>"




# =====================================================================
# RESOURCE LOADER
# =====================================================================
class ResourceLoader():

    def _drawAtlas(self, size, packedImgList):
        im1 = Image.new("RGBA",(size,size))
#        dr1 = ImageDraw.Draw(im1)
        for p in packedImgList:
            im1.paste(p.data, (p.x0,p.y0))
#            dr1.rectangle( (p.x0,p.y0,p.x1,p.y1), outline=(255,255,0,128) )
        return im1

    def _dumpData(self, boxList, dumpPath):
        objList = []
        i = 0
        for box in boxList:
            obj = {'id'  : i,
                   'name': box.name,
                   'x'   : box.x0,
                   'y'   : box.y0,
                   'w'   : box.width,
                   'h'   : box.height,
                   }
            objList.append(obj)
            i += 1
        fp = open(dumpPath, "wb")
        pickle.dump(objList, fp)
        fp.close()
    def _loadData(self, loadPath):
        fp = open(loadPath, "rb")
        objList = pickle.load(fp)
        fp.close()
        return objList

    def __init__(self, atlasDir="."):
        self._fonts         = []
        self._images        = {}
        self._textureByName = {}
        self._textureByID   = {}
        # Store atla dir
        self._atlasDir = atlasDir

    def addImage(self, name, imgPath, spriteInfo=None):
        id     = UniqueID.getID()
        imgRef = ResourceImage(name, imgPath, id, spriteInfo)
        #print(f"[LOADER][ADD] {imgRef}")
        if name in self._images:
            raise RuntimeError("ERROR: ResourceLoader - addImage - {name} is already registered")
        self._images[name] = imgRef

    # TODO do not regenerate font png file if already exists and not modified
    def addFont(self, name, fontPath, size=128, border=1, color=(255,255,255,255)):
        fontAtlasPath = f"{self._atlasDir}/font_{name}_{size}_{border}.png"
        fontDataPath  = f"{self._atlasDir}/font_{name}_{size}_{border}.dat"
        if not exists(fontAtlasPath) or not exists(fontDataPath):
            print(f"Regenerating {name} font with a size of {size} and a border of {border}...")
            # prepare char dimensions
            chars = {}
            # Open font using PIL
            fnt = ImageFont.truetype(fontPath, size)
            # For each ascii value (one byte length) check the width and height
            maxW = 0
            maxH = 0
            asciiMin = 32
            asciiMax = 128
            for ascii in range(asciiMin, asciiMax):
                v = ascii
                c = chr(ascii)
                # real size of the char
                dw, dh   = fnt.getsize(c)
                ofx, ofy = fnt.getoffset(c)
                chars[v] = {"char":c,
                            "size":size,
                            "width":dw,
                            "height":dh,
                            "offX":ofx,
                            "offY":ofy,
                            }
                maxW = max(maxW, dw)
                maxH = max(maxH, dh)
            maxW += 2*border
            maxH += 2*border
            # Here we have the maximum size of one character for the entire font
            # and we have a dictionary containing all character sizes
            # Create image and draw font characters
            N = 10
            W = N * maxW
            H = N * maxH
            im1 = Image.new("RGBA", (W,H), (0,0,0,0))
            dr1 = ImageDraw.Draw(im1)
            xRef = border
            yRef = border
            # prepare output
            OUT = {}
            for ascii in range(asciiMin, asciiMax):
                # Local data
                v   = ascii
                c   = chars[v]["char"]
                dw  = chars[v]["width"]
                dh  = chars[v]["height"]
                ofx = chars[v]["offX"]
                ofy = chars[v]["offY"]
                # Update position
                if xRef+dw+border >= W:
                    xRef  = border
                    yRef += maxH
                # Fill output
                OUT[v] = {"character":c,
                          "x":xRef,
                          "y":yRef,
                          "w":dw,
                          "h":dh}
                # Draw character at the right place
                # Y offset seems to ne used only to align characters correctly
                # So it seems useless to use it in the following code lines
                #dr1.rectangle((xRef,yRef,xRef+dw,yRef+dh),outline=color)
                #dr1.rectangle((xRef-border,yRef-border,xRef+dw+border,yRef+dh+border),outline=color)
                dr1 = ImageDraw.Draw(im1)
                dr1.text((xRef-ofx,yRef), c, font=fnt, fill=color)
                # Step to next estimated character position
                xRef += dw + 2*border
            # Save font image
            im1.save(fontAtlasPath)
            # Save font data
            fp = open(fontDataPath, "wb")
            pickle.dump(OUT, fp)
            fp.close()
        else:
            print(f"{name} font with a size of {size} and a border of {border} already exists")

        # Add font to list
        self._fonts.append(f"{name}_{size}_{border}")


    def generateImageAtlas(self, squareSize, maxSize, border, force=False):
        # Browse all images and get their hash, in order to create the atlas hash
        h = hashlib.sha256()
        # Use maxSize and border
        h.update(str(maxSize).encode())
        h.update(str(border).encode())
        # Browse all images
        for name in self._images:
            img = self._images[name]
            h.update(img.hash.encode())
        # Browse all fonts too
        for name in self._fonts:
            h.update(name.encode())
        d = h.hexdigest()
        atlasHash = d[16:48]
        # Create file paths
        atlasPath = f"{self._atlasDir}/atlas_{atlasHash}.png"
        dumpPath  = f"{self._atlasDir}/atlas_{atlasHash}.dat"

        # If the hash version of the atlas is not already generated,
        # We have to recreate it from scratch
        if not exists(atlasPath) or not exists(dumpPath) or force:
            print(f"\nGenerate new atlas (hash={atlasHash}) - side = {squareSize}")

            # Add images from fonts
            for name in self._fonts:
                # open data for this font
                fp = open(f"{self._atlasDir}/font_{name}.dat", "rb")
                fontData = pickle.load(fp)
                fp.close()
                # for each character, create an image
                for ch in fontData:
                    id = ch
                    info = fontData[ch]
                    charImg = ResourceImage(f"{name}_{id}", f"{self._atlasDir}/font_{name}.png", id, fontInfo=info)
                    self._images[f"{name}_{id}"] = charImg

            # First sort images by biggest width then biggest height
            lst = self._images.values()
            lst = sorted(lst, key=lambda x:(-x.spriteWidth,-x.spriteHeight))

            # For each image, make it fit into a squareSize structure
            binPacking   = BinPacking(squareSize, maxSize, border, lst)
            res, boxList, newSquareSize = binPacking.process()

            # Create atlasInfo Data (JSON format)
            self._dumpData(boxList, dumpPath)

            # when the structure is ready, use PILLOW to create the final atals
            # copy and paste images into it : Add a 2-pixel border with full transparency
            # in order to avoid texture bleeding during rendering
            im1 = self._drawAtlas(newSquareSize, boxList)
            im1.save(atlasPath)
            if not res:
                raise RuntimeError("Impossible to finish the Atlas generation !")
        else:
            print(f"\nRestore previous atlas (hash={atlasHash}) - size = {squareSize}")

        # Open json file and store atlas information
        textureList = self._loadData(dumpPath)
        for texture in textureList:
            id   = texture["id"]
            name = texture["name"]
            self._textureByID[id]     = texture
            self._textureByName[name] = texture

            print(f"{id}-{name}-{texture['x']}-{texture['y']}-{texture['w']}-{texture['h']}")

        self._atlasPath  = atlasPath


    def storeFonts(self, fsgpu):
        fonts = {}
        for f in self._fonts:
            minAscii = 32
            maxAscii = 127
            # Create block in the gpu
            # create it in the last part of the memory, as these data will not be modified,
            # it is better to put it far away from the modified areas (Sprites, ...)
            # because it will avoid updating data pages for nothing
            dataSize = maxAscii-minAscii+1
            blockID = fsgpu.alloc(dataSize, Gfx.TYPE_FONT, searchFromEnd=True)
            # Create array.array buffer in order to store font data
            data = array.array("f", [0.0, ] * int(dataSize))
            # store each texture ID related to each character ascii value
            i = 0
            for ascii in range(minAscii, maxAscii+1):
                name = f"{f}_{ascii}"
                charTex = self.getTextureByName(name)
                data[i] = charTex["id"]
                i += 1
            # copy the data into the FS GPU once for all
            fsgpu.write2Texture(blockID, data)
            # Store the blockID for this font
            fonts[f] = blockID
        # Store new list of fonts
        self._fonts = fonts

    def getNbTextures(self):
        if len(self._textureByName) != len(self._textureByID):
            print(self._textureByName,file=sys.stderr)
            print(self._textureByID, file=sys.stderr)
            raise RuntimeError("[ERROR] Difference between texture dicts !")
        return len(self._textureByName)

    def getTextureByName(self, name):
        if name not in self._textureByName:
            print(self._textureByName.keys())
            raise RuntimeError(f"[ERROR] name:{name} not in texture dict !")
        return self._textureByName[name]

    def getTextureById(self, id):
        if id not in self._textureByID:
            raise RuntimeError(f"[ERROR] id:{id} not in texture dict !")
        return self._textureByID[id]

    def getAllIds(self):
        return list(self._textureByID.keys())

    def getTextureAtlasPath(self):
        return self._atlasPath

    def getTextureImage(self):
        # dirPath  = Path(__file__).parent.resolve()
        filePath = self.getTextureAtlasPath()
        image = Image.open(filePath).convert("RGBA")
        image = image.transpose(Image.FLIP_TOP_BOTTOM)
        return image

    def getFontIDFromName(self, fontName):
        return self._fonts[fontName]




class BaseSystem():

    __slots__ = ["_compByRef",]

    # Recursive Dichotomy method to add the component at the correct place (sorted by priority)
    def _addComponent(self, ref, left, right):
        # Compute middle index
        mid = (left + right) // 2
        # Get index
        curP = self._compByRef[mid].getPriority()
        newP = ref.getPriority()
        # We reached the end of the dichotomy process
        if left == right:
            if newP > curP:
                left += 1
            # add the component add the left position
            self._compByRef.insert(left, ref)
            return
        else:
            if newP == curP:
                # Insert at the mid index
                self._compByRef.insert(mid, ref)
                return
            elif newP > curP:
                # move left border
                left += 1
                self._addComponent(ref, left, right)
            else:
                # move left border
                right -= 1
                self._addComponent(ref, left, right)


    def __init__(self):
        # Store components (by Ref, sorted by Priority)
        self._compByRef  = []

    def addComponent(self, ref):
        name = ref.getName()
        if "ollow" in name or "how" in name:
            print("insert "+name)

        if ref in self._compByRef:
            raise RuntimeError(f"[ERROR] cannot add the component {ref} twice in system {self} !")
        if len(self._compByRef) == 0:
            self._compByRef.append(ref)
        else:
            self._addComponent(ref, 0, len(self._compByRef)-1)

    def removeComponent(self, ref):
        if ref not in self._compByRef:
            raise RuntimeError(f"[ERROR] cannot remove the component {ref} from system {ref} !")
        self._compByRef.remove(ref)




import array


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
from .component import Component


class Gfx(Component):
# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

    # ------------------------------------
    #  CONSTANTS
    # ------------------------------------
    HEADER_SIZE    = 20
    TYPE_SPRITE    = 1
    TYPE_TEXT      = 2
    TYPE_RECTANGLE = 3
    TYPE_OVAL      = 4
    TYPE_FONT      = 5

    # ------------------------------------
    #  LOADER
    # ------------------------------------
    _loader = None
    _fsgpu  = None

    @staticmethod
    def setLoader(loader):
        Gfx._loader = loader

    @staticmethod
    def setFsGPU(fsgpu):
        Gfx._fsgpu = fsgpu


    # ------------------------------------
    #  SLOTS
    # ------------------------------------
    __slots__ = ['_blockID',
                 '_writeToFS',
                 '_data',
                ]

    # ------------------------------------
    #  CONSTRUCTOR
    # ------------------------------------
    def __init__(self,
                 width,
                 height,
                 scale=1.0,
                 x=0.0,
                 y=0.0,
                 z=0.0,
                 flipX=False,
                 flipY=False,
                 angle=0.0,
                 filterColor=(255,255,255),
                 visOn  = 1.0,
                 visTot = 1.0,
                 autoRotate=0.0,
                 anchorX=0.0,
                 anchorY=0.0,
                 dataSize=HEADER_SIZE,
                 gfxType=TYPE_SPRITE,
                 blockID=0,
                 name="Gfx",
                 ):
        # Call to mother class constructor
        super().__init__(Component.TYPE_GFX, name)

        # FS GPU data
        self._blockID   = blockID
        self._writeToFS = True
        self._data = array.array("f", [0.0, ] * int(dataSize))

        # id = 0-1-2-3 for R-G-B-A
        self.setColor(filterColor)
        # id = 4-5-6-7 for X-Y-W-H
        self.setX(x)
        self.setY(y)
        self.setW(width)
        self.setH(height)
        # id = 8-9-10-11 for SCALE-ANGLE-VISIBILITY ON/TOTAL
        self.setScale(scale)
        self.setAngle(angle)
        self.setVisibility(visOn, visTot)
        # id = 12 for angle auto-rotation
        self.setAutoRotate(autoRotate)
        # id = 13 for GFX type
        self._data[13] = gfxType
        # id = 14-15 for ANCHOR Horizontal-Vertical
        # The anchor is the sprite position origin (x/y), and from the rotation is performed
        # Both (anchorX and anchorY) are values in pixels
        # These values can exceed the sprite size : that means the anchor
        # point will be outside the texture rectangle area
        self.setAnchor(anchorX, anchorY)
        # id = 16 for Z-INDEX
        self.setZIndex(z)
        # id = 17-18 for FLIPX/FLIPY
        self.setFlipX(flipX)
        self.setFlipY(flipY)

    # ------------------------------------
    #  FS GPU
    # ------------------------------------
    def getBlockID(self):
        return self._blockID
    def getFsGpu(self):
        return Gfx._fsgpu
    def getGfxType(self):
        return self._data[13]

    # ------------------------------------
    #  POSITION (in pixels)
    # ------------------------------------
    def getX(self):
        return self._data[4]
    def getY(self):
        return self._data[4]
    def getPosition(self):
        return (self._data[4], self._data[5])
    def getTransform(self):
        return (self._data[4], self._data[5], self._data[9])
    def setX(self, v):
        self._data[4] = v
        self._writeToFS = True
    def setY(self, v):
        self._data[5] = v
        self._writeToFS = True
    def setPosition(self, x, y):
        self._data[4] = x
        self._data[5] = y
        self._writeToFS = True
    def moveX(self, dx):
        self._data[4] += dx
        self._writeToFS = True
    def moveY(self, dy):
        self._data[5] += dy
        self._writeToFS = True
    def movePosition(self, dx, dy):
        self._data[4] += dx
        self._data[5] += dy
        self._writeToFS = True
    def setTransform(self, x, y, ang):
        self._data[4] = x
        self._data[5] = y
        self._data[9] = ang
        self._writeToFS = True

    # ------------------------------------
    #  DIMENSIONS (in pixels, cannotbe modified directly)
    # ------------------------------------
    def getW(self):
        return self._data[6]
    def getH(self):
        return self._data[7]
    def setW(self, v):
        self._data[6] = v
        self._writeToFS = True
    def setH(self, v):
        self._data[7] = v
        self._writeToFS = True

    # ------------------------------------
    #  FILTER COLOR (0-255 values)
    # ------------------------------------
    def getColor(self):
        return (self._data[0], self._data[1], self._data[2], self._data[3])
    def setColor(self, v):
        alpha = 255
        if len(v) >= 4:
            alpha = v[3]
        self._data[0]   = v[0]
        self._data[1]   = v[1]
        self._data[2]   = v[2]
        self._data[3]   = alpha
        self._writeToFS = True

    # ------------------------------------
    #  SCALE
    # ------------------------------------
    def getScale(self):
        return self._data[8]
    def setScale(self, v):
        self._data[8] = v
        self._writeToFS = True

    # ------------------------------------
    #  ANGLE (in degrees)
    # ------------------------------------
    def getAngle(self):
        return self._data[9]
    def getAutoRotate(self):
        return self._data[12]
    def setAngle(self, v):
        self._data[9] = v
        self._writeToFS = True
    def setAutoRotate(self, v):
        self._data[12] = v
        self._writeToFS = True

    # ------------------------------------
    #  VISIBILITY (ON period and TOTAL period)
    # ------------------------------------
    def getVisibility(self):
        return (self._data[10], self._data[11])
    def setVisibility(self, on, total):
        self._data[10] = on
        self._data[11] = total
        self._writeToFS = True
    def show(self):
        self._data[10] = self._data[11]
        self._writeToFS = True
    def hide(self):
        self._data[10] = 0
        self._writeToFS = True

    # ------------------------------------
    #  ANCHOR
    # ------------------------------------
    def getAnchor(self):
        return (self._data[14], self._data[15])
    def getanchorX(self):
        return self._data[14]
    def getanchorY(self):
        return self._data[15]
    def setAnchor(self, x, y):
        self._data[14] = x
        self._data[15] = y
        self._writeToFS = True
    def setAnchorX(self, v):
        self._data[14] = v
        self._writeToFS = True
    def setAnchorY(self, v):
        self._data[15] = v
    def setAnchorLeft(self):
        self._data[14] = -1
        self._writeToFS = True
    def setAnchorCenterX(self):
        self._data[14] = 0
        self._writeToFS = True
    def setAnchorRight(self):
        self._data[14] = 1
        self._writeToFS = True
    def setAnchorTop(self):
        self._data[15] = -1
        self._writeToFS = True
    def setAnchorCenterY(self):
        self._data[15] = 0
        self._writeToFS = True
    def setAnchorBottom(self):
        self._data[15] = 1
        self._writeToFS = True
    def setAnchorCenter(self):
        self._data[14] = 0
        self._data[15] = 0
        self._writeToFS = True

    # ------------------------------------
    #  Z-INDEX
    # ------------------------------------
    def getZIndex(self):
        return self._data[16]
    def setZIndex(self, v):
        self._data[16] = v
        self._writeToFS = True
        # We have to notify the change of Z Index to the Gfx system
        # We get the entity and the scene to do it
        scn = self.getScene()
        if scn != None:
            scn.notifyChangeZ(self)

    # ------------------------------------
    #  FLIP SPRITE (X/Y)
    # ------------------------------------
    def getFlipX(self):
        return self._data[17]
    def getFlipY(self):
        return self._data[18]
    def setFlipX(self, v):
        v = 1.0 if v else 0.0
        self._data[17] = v
        self._writeToFS = True
    def setFlipY(self, v):
        v = 1.0 if v else 0.0
        self._data[18] = v
        self._writeToFS = True


    # ------------------------------------
    #  UPDATE (copy buffer into the GPU texture
    # ------------------------------------
    def update(self, deltaTime):
        # update data into FS if needed
        if self._writeToFS:
            #print(f"Writing {self._data} into FS")
            Gfx._fsgpu.write2Texture(self._blockID, self._data)
            self._writeToFS = False



# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class GfxSprite(Gfx):
# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

    # ------------------------------------
    #  SLOTS
    # ------------------------------------
    __slots__ = ["_blockID",
                 ]

    # ------------------------------------
    #  CONSTRUCTOR
    # ------------------------------------
    def __init__(self,
                 textureName,
                 width=-1, height=-1,
                 x=0.0,
                 y=0.0,
                 z=0.0,
                 flipX=False,
                 flipY=False,
                 scale=1.0,
                 angle=0.0,
                 visOn=1.0,
                 visTot=1.0,
                 autoRotate=0.0,
                 filterColor=(255,255,255),
                 anchorX=0.0,
                 anchorY=0.0,
                 name="GfxSprite"):
        NB_VALUES = Gfx.HEADER_SIZE + 4
        # Get texture info from loader
        texture   = Gfx._loader.getTextureByName(textureName)
        textureID = texture["id"]
        w         = texture["w"]
        h         = texture["h"]
        # print(f"Load GfxSprite : {textureID}-{textureName}-?-?-{w}-{h}")
        # if width or height is not filled, use the texture dimensions
        if width > 0:
            w = width
        if height > 0:
            h = height
        # Allocate buffer in the file system for it
        self._blockID = Gfx._fsgpu.alloc(NB_VALUES, Gfx.TYPE_SPRITE)
        # Call parent constructor
        super().__init__(w, h,
                         x=x,
                         y=y,
                         z=z,
                         flipX=flipX,
                         flipY=flipY,
                         angle=angle,
                         filterColor=filterColor,
                         scale=scale,
                         visOn=visOn,
                         visTot=visTot,
                         autoRotate=autoRotate,
                         anchorX=anchorX,
                         anchorY=anchorY,
                         dataSize=NB_VALUES,
                         gfxType=Gfx.TYPE_SPRITE,
                         blockID=self._blockID,
                         name=name+'_'+textureName)

        # Store specific information for this Sprite
        self.setTextureID(textureID)
        # Update the first time it is created
        self.update(1/60)

    # ------------------------------------
    #  TEXTURE ID
    # ------------------------------------
    def getTextureID(self):
        return self._data[Gfx.HEADER_SIZE + 0]
    def setTextureID(self, v):
        self._data[Gfx.HEADER_SIZE + 0]  = v
        self._writeToFS = True



# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class GfxBox(Gfx):
# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

    # ------------------------------------
    #  SLOTS
    # ------------------------------------
    __slots__ = ['_blockID',
                 ]

    # ------------------------------------
    #  CONSTRUCTOR
    # ------------------------------------
    def __init__(self,
                 inClr=(0,0,0,0),
                 width=-1,
                 height=-1,
                 x=0.0,
                 y=0.0,
                 z=0.0,
                 flipX=False,
                 flipY=False,
                 angle=0.0,
                 visOn=1.0,
                 visTot=1.0,
                 scale=1.0,
                 autoRotate=0.0,
                 filterColor=(255, 255, 255),
                 anchorX=0.0,
                 anchorY=0.0,
                 name="GfxBox"):

        # We need to add
        NB_VALUES = Gfx.HEADER_SIZE + 4
        # if width or height is not filled, use default dimensions
        w = 16
        h = 16
        if width > 0:
            w = width
        if height > 0:
            h = height
        # Allocate buffer in the file system for it
        self._blockID = Gfx._fsgpu.alloc(NB_VALUES, Gfx.TYPE_RECTANGLE)
        # Call parent constructor
        super().__init__(w, h,
                         x=x,
                         y=y,
                         z=z,
                         flipX=flipX,
                         flipY=flipY,
                         angle=angle,
                         filterColor=filterColor,
                         scale=scale,
                         visOn=visOn,
                         visTot=visTot,
                         autoRotate=autoRotate,
                         anchorX=anchorX,
                         anchorY=anchorY,
                         dataSize=NB_VALUES,
                         gfxType=Gfx.TYPE_RECTANGLE,
                         blockID=self._blockID,
                         name=name)

        # Store specific information for this Sprite (colors)
        self.setInColor(inClr)
        # Update the first time it is created
        self.update(1 / 60)

    # ------------------------------------
    #  INNER COLOR
    # ------------------------------------
    def getInColor(self):
        return (self._data[Gfx.HEADER_SIZE + 0],
                self._data[Gfx.HEADER_SIZE + 1],
                self._data[Gfx.HEADER_SIZE + 2],
                self._data[Gfx.HEADER_SIZE + 3])
    def setInColor(self, v):
        alpha = 255
        if len(v) >= 4:
            alpha = v[3]
        self._data[Gfx.HEADER_SIZE + 0]   = v[0]
        self._data[Gfx.HEADER_SIZE + 1]   = v[1]
        self._data[Gfx.HEADER_SIZE + 2]   = v[2]
        self._data[Gfx.HEADER_SIZE + 3]   = alpha
        self._writeToFS = True












#class GfxText(Gfx):
#
#    __slots__ = ['_blockID',
#                 '_fsgpu',
#                 '_message',
#                ]
#
#    def __init__(self, message, width=-1, height=-1, x=0.0, y=0.0, angle=0.0, scale=1.0, visOn=1.0, visTot=1.0,
#        autoRotate=0.0, filterColor=(255, 255, 255), fsgpu=None):
#        # We need to add specific data
#        NB_VALUES = Gfx.HEADER_SIZE + 4
#        # if width or height is not filled, use default dimensions
#        w = 16
#        h = 16
#        if width > 0:
#            w = width
#        if height > 0:
#            h = height
#        # Allocate buffer in the file system for it
#        self._blockID = fsgpu.alloc(NB_VALUES, Gfx.TYPE_TEXT)
#        # Call parent constructor
#        super().__init__(x, y,
#                         w, h,
#                         angle, scale,
#                         filterColor,
#                         visOn, visTot,
#                         autoRotate,
#                         fsgpu, NB_VALUES,
#                         Gfx.TYPE_RECTANGLE,
#                         self._blockID)
#        # Store specific information for this Sprite (colors)
#        self.setMessage(message)
#        # Update the first time it is created
#        self.update(1 / 60)
#
#    def setMessage(self, msg):
#        self._message = msg


#   def __init__(self, message, fontName, size=12, x=0.0, y=0.0, angle=0.0, scale=1.0, visOn=1.0, visTot=1.0, autoRotate=0.0, filterColor=(255,255,255), fsgpu=None):
#        # Specific Data for text element
#        # - font id (1)
#        # - font size (1)
#        # - message length (1)
#        # - message data (ascii) (L)
#        # TODO : if message data is too small to contain the new message, re-allocate a new block
#        # TODO : and free this one : do not forget to update the blockID property (add a setter for it ?)
#
#        # Get message length
#        L = len(message)
#        NB_VALUES = Gfx.HEADER_SIZE + 3 + L


#        # Get texture info from loader
#
#        # Allocate buffer in the file system for it
#        self._blockID = fsgpu.alloc(NB_VALUES, 1)     # TODO : 1 = TYPE SPRITE
#        # Call parent constructor
#        super().__init__(x, y, w, h, angle, scale, filterColor, visOn, visTot, autoRotate, fsgpu, NB_VALUES, self._blockID)
#        # Store specific information for this Sprite (id 10 for textureID)
#        self.setTextureID(textureID)
#        # Update the first time it is created
#        self.update(1/60)

#    def getTextureID(self):
#        return self._data[16]
#    def setTextureID(self, v):
#        self._data[16]  = v
#        self._writeToFS = True
#        self._fsgpu.writeBlock1(self._blockID, v, 10)
#    def __str__(self):
#        return f"<GfxSprite textureId={self.getTextureID()} blockID={self.getBlockID()} {super().__str__()}/>"









class SceneSystem():

    # Constructor
    def __init__(self):
        # Create the scene dictionary (name => ref)
        self._scenes       = {}
        # Create the special scene references
        self._currentScene = None
        self._nextScene    = None

    def run(self):
        pass

    # Scene management
    def registerScene(self, name, ref):
        # Check if this scene already exists
        if name in self._scenes:
            raise RuntimeError(f"[ERROR] The scene {name} is already registered !")
        # Add scene
        self._scenes[name] = ref
        # Make a link to the scene manager
        self._scenes[name].setManager(self)
        # Select this scene if this is the first to be added
        if len(self._scenes) == 1:
            self._currentScene = ref

    def unregisterScene(self, name):
        # Check if this scenes does not exist
        if name not in self._scenes:
            raise RuntimeError(f"[ERROR] Impossible to unregister the scene {name}. It does not exist !")
        # If this scene is the current selected one, raise an error
        if self._currentScene == self._scenes[name]:
            raise RuntimeError(f"[ERROR] Impossible to unregister the scene {name}. It is currently selected !")
        # Remove the link to the scene manager
        self._scenes[name].setManager(None)
        # Forget scene
        del self._scenes[name]

    def getNbScenes(self):
        return len(self._scenes)

    # Main application process
    def updateScene(self, deltaTime):
        if self._currentScene != None:
            # TODO | in case of transitions, do not update anymore ?
            # TODO | or update both scenes ?
            self._currentScene.updateWorld(deltaTime)

    def renderScene(self):
        if self._currentScene != None:
            # TODO | handle the transitions, rendering both scenes with shader, or plain rectangle
            self._currentScene.renderWorld()

    # Scene selection
    def selectScene(self, name):
        # Check if this scenes does not exist
        if name not in self._scenes:
            raise RuntimeError(f"[ERROR] Impossible to select the scene {name}. It does not exist !")
        # If the requested name is the selected one, do nothing
        if self._scenes[name] == self._currentScene:
            return
        # Now set the parameters to perform the transition
        self._nextScene = self._scenes[name]

    # Events
    def keyboardEvent(self, keyID, isPressed, modifiers):
        self._currentScene.keyboardEvent(keyID, isPressed, modifiers)
    def mouseButtonEvent(self, x, y, buttonID, isPressed, modifiers):
        self._currentScene.mouseButtonEvent(x, y, buttonID, isPressed, modifiers)
    def mouseMotionEvent(self, x, y, dx, dy):
        self._currentScene.mouseMotionEvent(x, y, dx, dy)
    def mouseDragEvent(self, x, y, dx, dy, buttonID, modifiers):
        self._currentScene.mouseDragEvent(x, y, dx, dy, buttonID, modifiers)
    def mouseScrollEvent(self, x, y, dx, dy):
        self._currentScene.mouseScrollEvent(x, y, dx, dy)
    def gamepadButtonEvent(self, gamepadID, buttonID, isPressed):
        self._currentScene.gamepadButtonEvent(gamepadID, buttonID, isPressed)
    def gamepadAxisEvent(self, gamepadID, axisID, analogValue):
        self._currentScene.gamepadAxisEvent(gamepadID, axisID, analogValue)











class Buttons():

    ANY    = -1
    A      =  0
    B      =  1
    X      =  2
    Y      =  3
    LB     =  4
    RB     =  5
    VIEW   =  6
    MENU   =  7
    LSTICK =  8
    RSTICK =  9

    MOUSE_LEFT   = 1
    MOUSE_CENTER = 2
    MOUSE_RIGHT  = 4


class Gamepads():

    ANY = -1

class Axis():

    ANY = -1
    LX = "x"
    LY = "y"
    RX = "rx"
    RY = "ry"
    Z  = "z"




from array import array


class GfxSystem():

    _glData = None

    @staticmethod
    def setOpenGlData(glData):
        GfxSystem._glData= glData

    # Recursive Dichotomy method to add the component at the correct place (sorted by Z)
    def _addComponent(self, ref, left, right):
        # We reached the end of the dichotomy process
        if right - left == -1:
            # We went too far, take the "left" index
            right = left
        # Compute middle index
        mid = (left + right) // 2
        # Get index
        curZ = self._compByRef[mid].getZIndex()
        newZ = ref.getZIndex()
        if left == right:
            if newZ > curZ:
                left += 1
            # add the component add the left position
            self._compByRef.insert(left, ref)
            return
        else:
            if newZ == curZ:
                # Insert at the mid index
                self._compByRef.insert(mid, ref)
                return
            elif newZ > curZ:
                # move left border
                left = mid+1
                self._addComponent(ref, left, right)
            else:
                # move left border
                right = mid-1
                self._addComponent(ref, left, right)

    def _genVertexBuffer(self):
        for c in self._compByRef:
            yield c.getBlockID()

    def __init__(self):
        # Store components (by Ref, sorted by Z Index)
        self._compByRef  = []

    def addComponent(self, ref):
        #print(f"Adding {ref.getName()} @ Z={ref.getZIndex()} ...")
        if ref in self._compByRef:
            raise RuntimeError(f"[ERROR] cannot add the Gfx {ref} twice !")
        if len(self._compByRef) == 0:
            self._compByRef.append(ref)
        else:
            self._addComponent(ref, 0, len(self._compByRef)-1)

    def removeComponent(self, ref):
        if ref not in self._compByRef:
            raise RuntimeError(f"[ERROR] cannot remove the Gfx {ref} !")
        self._compByRef.remove(ref)

    def notifyChangeZ(self, ref):
        # Remove the component and add it once again
        self.removeComponent(ref)
        self.addComponent(ref)

    def update(self, deltaTime, systemTime):
        # Update all the gfx components
        for c in self._compByRef:
            c.update(deltaTime)
        # Update the GPU file system
        GfxSystem._glData.update(deltaTime, systemTime)

    def render(self):
        vb = array("l", self._genVertexBuffer())
        GfxSystem._glData.updateVertexBuffer( vb, len(self._compByRef) )
        GfxSystem._glData.render()



from .base_system import BaseSystem

class ScriptSystem(BaseSystem):

    def __init__(self):
        super().__init__()

    def notifyChangeScriptPriority(self, ref):
        # Remove the component and add it once again
        self.removeComponent(ref)
        self.addComponent(ref)

    def updateScripts(self, deltaTime, systemTime):
        # Update all the Script components
        for c in self._compByRef:
            c.updateScript(deltaTime, systemTime)



from .base_system  import BaseSystem
from ..main.common import Buttons, Gamepads, Axis


class KeyboardSystem(BaseSystem):

    def __init__(self):
        super().__init__()

    def keyboardEvent(self, keyID, isPressed, modifiers):
        # Update all the Script
        for c in self._compByRef:
            ID = f"K-{keyID}"
            c.notifyEvent(ID, isPressed, modifiers)
        #print(f"\n\n\nnotified {len(self._compByRef)} input components\n\n\n")


class GamepadButtonSystem(BaseSystem):

    def __init__(self):
        super().__init__()

    def gamepadButtonEvent(self, gamepadID, buttonID, isPressed):
        print(f"Pad:{gamepadID} - button:{buttonID} - isPressed:{isPressed}")
        # Update all the Script components
        for c in self._compByRef:
            # Notify specific event
            ID = f"G-{gamepadID}-{buttonID}"
            c.notifyEvent(ID, isPressed)
            # Notify for current button and any gamepad
            ID = f"G-{Gamepads.ANY}-{buttonID}"
            c.notifyEvent(ID, isPressed)
            # Notify for current gamepad and any button
            ID = f"G-{gamepadID}-{Buttons.ANY}"
            c.notifyEvent(ID, isPressed)
            # Notify for any gamepad and any button
            ID = f"G-{Gamepads.ANY}-{Buttons.ANY}"
            c.notifyEvent(ID, isPressed)



class GamepadAxisSystem(BaseSystem):

    def __init__(self):
        super().__init__()

    def gamepadAxisEvent(self, gamepadID, axisID, value):
        # print(f"Pad:{gamepadID} - button:{axisID} - value:{value}")
        # Update all the Script components
        for c in self._compByRef:
            # Notify specific event
            ID = f"A-{gamepadID}-{axisID}"
            c.notifyEvent(ID, value)
            # Notify any gamepad
            ID = f"A-{Gamepads.ANY}-{axisID}"
            c.notifyEvent(ID, value)
            # Notify any axis
            ID = f"A-{gamepadID}-{Axis.ANY}"
            c.notifyEvent(ID, value)
            # Notify any axis on any gamepad
            ID = f"A-{Gamepads.ANY}-{Axis.ANY}"
            c.notifyEvent(ID, value)
            # Notify any axis on any gamepad




class MouseSystem(BaseSystem):

    def __init__(self):
        super().__init__()

    def mouseButtonEvent(self, x, y, buttonID, isPressed, modifiers):
        # Update all the Script components
        for c in self._compByRef:
            # Notify for this specific button
            ID = f"M-{buttonID}"
            c.notifyEvent(ID, isPressed, modifiers, x, y)
            # Notify for any button
            ID = f"M-{Buttons.ANY}"
            c.notifyEvent(ID, isPressed, modifiers, x, y)





import math

from ....ecs3.components.script import Script


class MoveCircle(Script):

    def __init__(self, gfx, center, radius, speed, keyboard, name="MoveCircle"):
        super().__init__(1, name)
        self._gfx    = gfx
        self._center = [ center[0], center[1] ]
        self._radius = radius
        self._speed  = speed    # degrees per second
        self._key    = keyboard

    def updateScript(self, deltaTime, systemTime):
        # Move center according to pressed keys
        if self._key.isPressed("moveUp"):
            self._center[1] -= 20 * 60 * deltaTime
        if self._key.isPressed("moveDown"):
            self._center[1] += 20 * 60 * deltaTime

        # Set the position according to the speed and system time
        x = math.cos(self._speed*systemTime*2*math.pi/360) * self._radius + self._center[0]
        y = math.sin(self._speed*systemTime*2*math.pi/360) * self._radius + self._center[1]

        self._gfx.setX(x)
        self._gfx.setY(y)



class MoveSquare(Script):

    def __init__(self, gfx, center, radius, speed, keyboard, name="MoveCircle"):
        super().__init__(1, name)
        self._gfx    = gfx
        self._center = [ center[0], center[1] ]
        self._radius = radius
        self._speed  = speed    # degrees per second
        self._key    = keyboard

    def updateScript(self, deltaTime, systemTime):
        # Move center according to pressed keys
        if self._key.isPressed("moveLeft"):
            self._center[0] -= 20 * 60 * deltaTime
        if self._key.isPressed("moveRight"):
            self._center[0] += 20 * 60 * deltaTime

        # Set the position according to the speed and system time
        ratio = self._speed*systemTime/360
        ratio = ratio - int(ratio)
        if ratio <0.25:
            x = 2*self._radius * ratio *4 - self._radius + self._center[0]
            y = - self._radius + self._center[1]
        elif ratio < 0.5:
            y = 2 * self._radius * (ratio-0.25) * 4 - self._radius + self._center[1]
            x = self._radius + self._center[0]
        elif ratio < 0.75:
            y = self._radius + self._center[1]
            x = - 2*self._radius * (ratio-0.5) *4 + self._radius + self._center[0]
        else:
            x = - self._radius + self._center[0]
            y = - 2 * self._radius * (ratio-0.75) * 4 + self._radius + self._center[1]

        self._gfx.setX(x)
        self._gfx.setY(y)


class ShowHide(Script):

    def __init__(self, gfx, pad, mouse, name="ShowHide"):
        super().__init__(0, name)
        self._gfx   = gfx
        self._pad   = pad
        self._mouse = mouse

    def updateScript(self, deltaTime, systemTime):
        # Gamepad buttons
        if self._pad.hasBeenPressed("Hide"):
            self._gfx.hide()
        if self._pad.hasBeenPressed("Show"):
            self._gfx.show()
        if self._pad.hasBeenPressed("Enlarge"):
            scale = self._gfx.getScale()
            self._gfx.setScale(scale * 1.5)
        if self._pad.hasBeenPressed("Reduce"):
            scale = self._gfx.getScale()
            self._gfx.setScale(scale / 1.5)
        # Mouse buttons
        if self._mouse.isPressed("RotateLeft"):
            angle = self._gfx.getAngle()
            self._gfx.setAngle(angle - 15)
        if self._mouse.isPressed("RotateRight"):
            angle = self._gfx.getAngle()
            self._gfx.setAngle(angle + 15)


class MoveShip(Script):

    def __init__(self, gfx, axis, name="MoveShip"):
        super().__init__(0, name)
        self._gfx  = gfx
        self._axis = axis

    def updateScript(self, deltaTime, systemTime):
        # Compute and set angle
        x = self._axis.getAnalogValue("moveX")
        y = self._axis.getAnalogValue("moveY")
        ang = math.atan2(y,x)*180.0/math.pi
        self._gfx.setAngle(ang+90)
        # update position
        x *= 10
        y *= 10
        self._gfx.movePosition(x,y)



import math
import time
from random import choice, randint

from .fsgpu_buffer import FsGpuBuffer


class FsGpuMain():

    USE_LOCAL_FS_BUFFER = False

    __slots__ = ['_pageShift',
                 '_maxPages',
                 '_pageSize',
                 '_pageMask',
                 '_nbPages',
                 '_nbComp',
                 '_ctx',
                 '_texture',
                 '_pages',
                ]

    # ----------------------------------------------------
    # CONSTRUCTOR
    # ----------------------------------------------------
    def __init__(self, ctx, pageSize, nbPages):
        # pageShift is the number of bits to code the pageSize
        # 16 means the page size can be 64kB
        # We store the page number on the 16 remaining bits (to stay in unsigned 32 bits range)
        self._pageShift = 16
        self._pageMask  = (1 << (32 - self._pageShift)) - 1
        self._maxPages  = math.pow(2, 32-self._pageShift)
        # Either page size cannot be more than 2^pageShift
        if pageSize > math.pow(2, self._pageShift):
            raise RuntimeError(f"[ERROR] bad page size : size={pageSize} is too big. Do not use more than {math.pow(2, self._pageShift)} !")
        if nbPages > self._maxPages:
            raise RuntimeError(f"[ERROR] bad NB pages : nbPages={nbPages} maxPages={self._maxPages} pageSize={pageSize} !")

        # Each page is a FsGpuBuffer instance
        self._pageSize = pageSize
        self._nbPages  = nbPages
        self._nbComp   = 4
        # Store gl context
        self._ctx = ctx
        # init buffers and texture
        self._clear()


    # ----------------------------------------------------
    # PROPERTIES
    # ----------------------------------------------------
    def getMaxPages(self):
        return self._maxPages
    def getNbPages(self):
        return self._nbPages
    def getPageSize(self):
        return self._pageSize
    def getTotalSize(self):
        return self._pageSize * self._nbPages
    def getTexture(self):
        return self._texture


    # ----------------------------------------------------
    # PRIVATE METHODS
    # ----------------------------------------------------
    def _clear(self):
        # create buffers
        self._pages = [FsGpuBuffer(self._pageSize) for N in range(self._nbPages)]
        # Create texture from context
        self._texture = self._ctx.texture((self._pageSize, self._nbPages), self._nbComp, dtype="f4")

    def _isPageOK(self, pageNum):
        return pageNum >= 0 and pageNum < self._nbPages

    def _createID(self, offset, page):
        return int((page << self._pageShift) + offset)

    def _explodeID(self, id):
        offMask  = (1 << self._pageShift) - 1
        offset   = id & offMask
        page     = (id >> self._pageShift) & self._pageMask
        return (offset, page)

    # ----------------------------------------------------
    # PUBLIC API
    # ----------------------------------------------------
    # the ascending parameter is used to go through the pages
    # in order to find a valid available space (from tge beginning or the end)
    # The page number AND the buffer offset are gathered and returned as a buffer ID
    def alloc(self, userSize, userType, searchFromEnd=False, data=None):
        # Init vars
        step   = 1
        offset = None
        page = 0
        if searchFromEnd:
            step = -1
            page = len(self._pages)-1
        # Process allocation
        while offset == None:
            # check current page number
            if not self._isPageOK(page):
                dmp = self.dump()
                raise RuntimeError(f"[ERROR] page number is not correct during allocation. May be the memory is full, or the requested start page is not correct. page={page} !\n" + dmp)
            # Try to allocate a block into the current page (can return None if no available space)
            offset = self._pages[page].alloc(userSize, userType)
            # Go to next page (if needed)
            page += step
        # Here we have a valid offset, get block ID (page must be -1 because of the loop)
        id = self._createID(offset, page - step)
        # Fill the buffer if requested
        if data != None:
            self.write2Texture(id, data)
        # return block ID
        return id

    def free(self, blockID):
        # retrieve block position information
        offset, page = self._explodeID(blockID)
        if not self._isPageOK(page):
            raise RuntimeError(f"[ERROR] bad page value from block ID ! id={blockID} - page={page} - offset={offset}")
        # Now free this block from the memory
        self._pages[page].free(offset)


    def readFromTexture(self, id):
        # retrieve block position information
        offset, page = self._explodeID(id)
        # Get data from texture (GPU)
        gpuData = 0 # TODO : retrieve data from texture directly (is this method useful ?
                    # TODO : as there is alreasdy the Sprite instances that contain everything ?
        # Get block data (local CPU buffer)
        if FsGpuMain.USE_LOCAL_FS_BUFFER :
            cpuData = self._pages[page].read(offset)
            # compare gpuData and cpuData to be sure there is no issue somewhere in the dataflow
            # TODO

    def write2Texture(self, id, data):
        # retrieve block position information
        offset, page = self._explodeID(id)
        # Write into the CPU array.array (that is a CPU copy of the file system)
        if FsGpuMain.USE_LOCAL_FS_BUFFER :
            self._pages[page].write(offset, data)
        # write block data directly to texture
        offset = (offset + FsGpuBuffer.OVERHEAD)//self._nbComp
        self._texture.write(data, viewport=(offset, page, len(data)//self._nbComp, 1))


    # ----------------------------------------------------
    # APP PROCESS
    # ----------------------------------------------------
    def update(self, deltaTime):
        pass

        # ~~~~~~~~~~~~~~~~~~~~~~~~
        # We will call the page buffer defrag process, and according
        # to the frame time, we will see how many pages we can defrag in-a-row
        # ~~~~~~~~~~~~~~~~~~~~~~~~
        # TODO : For the moment we only defrag one page
        # TODO : improve defrag process (or algorithm underneath) because it drops the perfs !!
#        for page in self._pages:
#            res = page.defrag()
#            if res:
#                break

        # ~~~~~~~~~~~~~~~~~~~~~~~~
        # Browse all buffers and check their 'modified' property
        # copy the buffer data into the texture and call resetModify()
        # when finished
        # ~~~~~~~~~~~~~~~~~~~~~~~~
        # As there is a way to fill data from the bottom page of the FS
        # this could be efficient to put all static blocks at the end of the
        # memory, and the static ones at the beginning

        # TODO : may be force a page to allocate blocks, in order to handle
        # which pages are willing to be modified or not
        # TODO : this process can be improved by getting the small parts
        #        of the buffers that have been modified, instead of rewriting
        #        the whole buffer
#        i = 0
#        while i < self._nbPages:
#            p = self._pages[i]
#            if p.isModified():
#                # print(f"[FS GPU] Writing page #{i} into the GPU texture")
#                # write this page into the texture
#                self._texture.write(p.getData(), viewport=(0, i, self._pageSize, 1))
#                # buffer has been updated into the texture
#                # reset flag
#                p.resetModify()
#            # print(f"PAGE WRITE = {lap2-lap1}")
#            i += 1

    def render(self):
        pass


    # ----------------------------------------------------
    # DEBUG
    # ----------------------------------------------------
    def display(self, displayData=False):
        print(self.dump(displayData))

    def dump(self, displayData=False):
        out = ""
        i = 0
        for page in self._pages:
            out +=  "=================================================" + "\n"
            out +=  "=================    PAGE %02d    =================" % (i)  + "\n"
            out += f"================= modified={page.isModified()} =================" + "\n"
            out +=  "=================================================" + "\n"
            i += 1
            out += page.dump(displayData)

        out += "\n"
        return out


    # ----------------------------------------------------
    # TESTS
    # ----------------------------------------------------
    def test002(self):
        TEST_NUMS = [10,20,50,100,200,500,1000,2000,5000,10000,20000]
        BSIZE = 8
        TYPE  = 1
        data = [10 + i for i in range(BSIZE)]
        results = []
        results.append("Nalloc\tTalloc\tNfree\tTfree")
        for T in TEST_NUMS:
            # Init buffers
            allocTime = []
            freeTime = []
            blockIDs = []
            self._clear()

            # Perform tests
            print("------------------------------------")
            for N in range(T):
                # Defrag FS
                self.update(1/60)

                # Try to free up some buffers
                if len(blockIDs) > 10:
                    nbRemove = randint(1,len(blockIDs)//10)
                    for i in range(nbRemove):
                        # Free block
                        idRemove = choice(blockIDs)
                        time0 = time.time()
                        self.free(idRemove)
                        time1 = time.time()
                        # Add allocation time
                        freeTime.append(time1 - time0)
                        # Remove id from list
                        blockIDs.remove(idRemove)

                # Try to allocate some blocks
                nbAdd = randint(1, 50)
                for i in range(nbAdd):
                    # Allocate block
                    time0 = time.time()
                    id = self.alloc(BSIZE, TYPE, data=data)
                    time1 = time.time()
                    # Add allocation time
                    allocTime.append(time1-time0)
                    # Store block ID
                    blockIDs.append(id)

            # display FS
            #self.display()

            res = ""
            # Compute min, max and average of times
            summ = round(1000 * sum(allocTime), 2)
            mini = round(1000 * min(allocTime), 2)
            maxi = round(1000 * max(allocTime), 2)
            avrg = round(1000 * sum(allocTime)/len(allocTime), 2)
            res += f"{len(allocTime)}\t{summ}\t"
            print(f"ALLOC : N={len(allocTime)}")
            print(f"    sum={summ}ms mini={mini}ms maxi={maxi}ms average={avrg}ms")
            # Compute min, max and average of times
            summ = round(1000 * sum(freeTime), 2)
            mini = round(1000 * min(freeTime), 2)
            maxi = round(1000 * max(freeTime), 2)
            avrg = round(1000 * sum(freeTime)/len(freeTime), 2)
            res += f"{len(freeTime)}\t{summ}\t"
            print(f"FREE  : N={len(freeTime)}")
            print(f"    sum={summ} mini={mini}ms maxi={maxi}ms average={avrg}ms")
            res = res.replace(".",",")
            results.append(res)

        print()
        print("====================================")
        for r in results:
            print(r)


from .component import Component


# ============================================================
# Storage of any action status
# ============================================================
class LogicStatus():

    __slots__ = ["_state",
                 "_risingEdge",
                 "_fallingEdge",
                 "_modifiers",
                 "_lastX",
                 "_lastY",
                 ]

    def __init__(self):
        self._state       = False
        self._risingEdge  = False
        self._fallingEdge = False
        self._modifiers   = None
        self._lastX       = 0
        self._lastY       = 0

    def updateStatus(self, v, modifiers=None, x=0, y=0):
        # If there is a change
        if v != self._state:
            if v:
                self._risingEdge = True
            else:
                self._fallingEdge = True
            self._state = v
            self._modifiers = modifiers
            self._lastX = x
            self._lastY = y

    def isPressed(self):
        return self._state
    def isReleased(self):
        return not self._state
    def hasBeenPressed(self):
        res = self._risingEdge
        self._risingEdge = False
        return res
    def hasBeenReleased(self):
        res = self._fallingEdge
        self._fallingEdge = False
        return res
    def getModifiers(self):
        return self._modifiers
    def getLastX(self):
        return self._lastX
    def getLastY(self):
        return self._lastY


# ============================================================
# Management of a logic input
# ============================================================
class LogicButton(Component):

    __slots__ = ["_buttons",
                 "_actions",
                 ]

    def __init__(self, type, name="LogicButton"):
        super().__init__(type, name)
        # Contains the action names for each key id
        self._buttons    = {}
        # Contains the status for each action
        self._actions = {}

    def addAction(self, ID, actionName):
        # Create action if not already present
        if actionName not in self._actions:
            self._actions[actionName] = LogicStatus()
        # Add key
        if ID not in self._buttons:
            self._buttons[ID] = []
        # Add key-action if needed
        if actionName not in self._buttons[ID]:
            self._buttons[ID].append(actionName)

    def removeAction(self, ID, actionName):
        # TODO
        pass

    # ID : "K" + keyID for keyboard
    #      "G" + gamepadID + buttonID for gamepads
    #      "M" + buttonID for mouse
    def notifyEvent(self, ID, isPressed, modifiers=None, x=0, y=0):
        if ID in self._buttons:
            names = self._buttons[ID]
            for name in names:
                self._actions[name].updateStatus(isPressed, modifiers, x, y)

    def isPressed(self, actionName):
        res = False
        if actionName in self._actions:
            res = self._actions[actionName].isPressed()
        return res

    def isReleased(self, actionName):
        res = False
        if actionName in self._actions:
            res = self._actions[actionName].isReleased()
        return res

    def hasBeenPressed(self, actionName):
        res = False
        if actionName in self._actions:
            res = self._actions[actionName].hasBeenPressed()
        return res

    def hasBeenReleased(self, actionName):
        res = False
        if actionName in self._actions:
            res = self._actions[actionName].hasBeenReleased()
        return res


# ============================================================
# KEYS
# ============================================================
class Keyboard(LogicButton):

    __slots__ = []

    def __init__(self, name="Keyboard"):
        super().__init__(Component.TYPE_KEY, name)

    def addKey(self, keyID, actionName):
        ID = f"K-{keyID}"
        super().addAction(ID, actionName)

    def removeKey(self, keyID, actionName):
        # TODO
        pass


# ============================================================
# GAMEPAD BUTTONS
# ============================================================
class GamepadButton(LogicButton):

    __slots__ = []

    def __init__(self, name="GamepadButton"):
        super().__init__(Component.TYPE_PAD_BUTTON, name)

    def addButton(self, gamepadID, buttonID, actionName):
        ID = f"G-{gamepadID}-{buttonID}"
        super().addAction(ID, actionName)

    def removeButton(self, gamepadID, buttonID, actionName):
        # TODO
        pass


# ============================================================
# MOUSE BUTTONS
# ============================================================
class MouseButton(LogicButton):

    __slots__ = []

    def __init__(self, name="MouseButton"):
        super().__init__(Component.TYPE_MOUSE_BUTTON, name)

    def addButton(self, buttonID, actionName):
        ID = f"M-{buttonID}"
        super().addAction(ID, actionName)

    def removeButton(self, buttonID, actionName):
        # TODO
        pass










# ============================================================
# Management of a logic input
# ============================================================
class GamepadAxis(Component):

    __slots__ = ["_axis",
                 "_actions",
                 "_dead",
                 ]

    def __init__(self, deadZone=0.2, name="GamepadAxis"):
        super().__init__(Component.TYPE_PAD_AXIS, name)
        # Contains the action names for each key id
        self._axis = {}
        # Contains the status for each action
        self._actions = {}
        # Dead zone value
        self._dead = deadZone

    def addAxis(self, gamepadID, axisID, actionName):
        # Build ID
        ID = f"A-{gamepadID}-{axisID}"
        # Create action if not already present
        if actionName not in self._actions:
            self._actions[actionName] = 0.0
        # Add ID
        if ID not in self._axis:
            self._axis[ID] = []
        # Add key-action if needed
        if actionName not in self._axis[ID]:
            self._axis[ID].append(actionName)

    def removeAction(self, gamepadID, axisID, actionName):
        # TODO
        pass

    def notifyEvent(self, ID, analogValue):
        # Saturate value
        if abs(analogValue) < self._dead:
            analogValue = 0.0
        if ID in self._axis:
            names = self._axis[ID]
            for name in names:
                self._actions[name] = analogValue

    def getAnalogValue(self, actionName):
        res = 0.0
        if actionName in self._actions:
            res = self._actions[actionName]
        return res





from .default_shader import Shader

class SimpleShader(Shader):

    def __init__(self, name="Shader v1"):
        super().__init__(name)

    def _getHeader(self):
        return """
            #version 330
            
            // ------------ FILE SYSTEM ----------------
            #define PAGE_SIZE_BITS (16)
            #define PAGE_NUM_BITS  (16)            
            #define OFFSET_MASK    ((1<<PAGE_SIZE_BITS)-1)
            #define PAGE_MASK      ( ((1<<PAGE_NUM_BITS )-1) << PAGE_SIZE_BITS )
            #define OVERHEAD       (1)

            // ------------ GFX TYPES ----------------
            #define TYPE_SPRITE    (1)
            #define TYPE_TEXT      (2)
            #define TYPE_RECTANGLE (3)
            #define TYPE_OVAL      (4)
            #define TYPE_FONT      (5)
            

        """

    def getVertex(self):
        return self._getHeader() + """
            // The per sprite input data
            in int blockID;
            
            // Uniforms
            uniform sampler2D   fsGpuChan;
                
            // Output                
            out ivec2 fsCoords;

            void main() {
            
                // Extract X,Y texture position from offset
                int fsOffset = (blockID & OFFSET_MASK) >> 2;
                int fsPage   = (blockID & PAGE_MASK  ) >> PAGE_SIZE_BITS;
                fsOffset    += OVERHEAD;
                fsCoords     = ivec2(fsOffset,fsPage);
                
            }
        """

    def getGeometry(self):
        return self._getHeader() + """
            // We are taking single points form the vertex shader
            // and emitting 4 new vertices creating a quad/sprites
            layout (points) in;
            layout (triangle_strip, max_vertices = 4) out;

            uniform mat4        projection;
            uniform sampler2D   atlasTextureChan;
            uniform sampler2D   fsGpuChan;
            uniform usampler2D  atlasInfoChan;
            uniform float       systemTime;

            // Since geometry shader can take multiple values from a vertex
            // shader we need to define the inputs from it as arrays.
            // In our instance we just take single values (points)
            in ivec2 fsCoords[];

            out vec2 uv;
            out vec4 inColor;
            out vec4 filterColor;
            out float gfxType;

            void processSprite(){
            
            }
            
            void processBox(){
            
            }

            void main() {

                // Get coord in the FS texture                
                ivec2 fsTexelCoords = fsCoords[0];
                
                //-------------------------------------------------------------------
                // Get GFX Header data
                //-------------------------------------------------------------------

                // Get filtering color
                filterColor = texelFetch( fsGpuChan, fsTexelCoords, 0 )/255.0;
                // Init IN and OUT colors with the same values
                inColor  = filterColor;
                fsTexelCoords.x  += 1;

                // Get position and size
                vec2 posFS  = texelFetch( fsGpuChan, fsTexelCoords, 0 ).xy;
                vec2 sizeFS = texelFetch( fsGpuChan, fsTexelCoords, 0 ).zw;
                fsTexelCoords.x += 1;
                
                // Get Scale / angle / visibility values
                float scaleFS    = texelFetch( fsGpuChan, fsTexelCoords, 0 ).x;
                float angleFS    = texelFetch( fsGpuChan, fsTexelCoords, 0 ).y;
                float fsVisOn    = texelFetch( fsGpuChan, fsTexelCoords, 0 ).z;
                float fsVisTotal = texelFetch( fsGpuChan, fsTexelCoords, 0 ).w;
                fsTexelCoords.x += 1;

                // Get Auto-Rotate / gfx type / Anchor X-Y
                float fsAutoRot  = texelFetch( fsGpuChan, fsTexelCoords, 0 ).x;
                float fsType     = texelFetch( fsGpuChan, fsTexelCoords, 0 ).y;
                vec2  fsAnchor   = texelFetch( fsGpuChan, fsTexelCoords, 0 ).zw;
                fsTexelCoords.x += 1;

                // Get Z-Index
                float fsZIndex = texelFetch( fsGpuChan, fsTexelCoords, 0 ).x;
                // Get FLIPX and FLIPY
                vec2 fsFlip = texelFetch( fsGpuChan, fsTexelCoords, 0 ).yz;
                // TODO : 1 remaining data for future use
                fsTexelCoords.x += 1;


                //-------------------------------------------------------------------
                // CHECK if this Gfx is visible or not
                //-------------------------------------------------------------------

                // Hide element if needed                
                if (fract(systemTime/fsVisTotal) > (fsVisOn/fsVisTotal)){
                    return;
                }
                
                // Set center and halfsize of sprite
                vec2 center = posFS;
                vec2 hsize  = scaleFS * sizeFS / 2.0;

                // Prepare data for geometry process
                vec2 pos0 = vec2(0.0);
                vec2 pos1 = vec2(0.0);
                


                //==========================================================
                // SPRITE
                //==========================================================
                if (fsType == TYPE_SPRITE){
                    // Get texture ID (for SPRITE)
                    float textIdFS = texelFetch( fsGpuChan, fsTexelCoords, 0 ).x;
                    fsTexelCoords.x += 1;
                    
                    // Get whole Atlas dimensions
                    vec2  texDim  = textureSize(atlasTextureChan,0);
                    // Get position and size of texture from the atlas
                    uvec4 texBox2 = texelFetch(atlasInfoChan, ivec2(int(textIdFS),0), 0 );
                    // Set values from unsigned to signed
                    vec4  texBox  = vec4(texBox2);
                    // Half pixel
                    vec2 halfPixel = vec2(0.0)/texDim;   // set 0.5 for half pixel feature
    
                    // Get texture coords
                    pos0   = (texBox.xy/texDim.xy);
                    pos1   = pos0 + (texBox.zw/texDim.xy);
                    pos0   = pos0 + halfPixel;
                    pos1   = pos1 - halfPixel;
                    pos0.y = 1.0-pos0.y;
                    pos1.y = 1.0-pos1.y;
                    // Flip texture if needed
                    float tmp;
                    if (fsFlip.x != 0.0){
                        tmp    = pos0.x;
                        pos0.x = pos1.x;
                        pos1.x = tmp;
                    }
                    if (fsFlip.y != 0.0){
                        tmp    = pos0.y;
                        pos0.y = pos1.y;
                        pos1.y = tmp;
                    }
                }
                
                //==========================================================
                // BASIC BOX SHAPE
                //==========================================================
                if (fsType == TYPE_RECTANGLE){
                    // get inside color
                    vec4 fsInClr = texelFetch( fsGpuChan, fsTexelCoords, 0 )/255.0;
                    fsTexelCoords.x += 1;    
                    inColor  = fsInClr;
                }


                //-------------------------------------------------------------------
                // CONVERT size and position according to current viewport
                //-------------------------------------------------------------------

                // Get center position and check if it is outside viewport
                vec4 glMin = projection * vec4(center-hsize, 0.0, 1.0);
                vec4 glMax = projection * vec4(center+hsize, 0.0, 1.0);
                if ( (abs(glMin.x) > 1.0 && abs(glMax.x) > 1.0)
                     || 
                     (abs(glMin.y) > 1.0 && abs(glMax.y) > 1.0)
                    ){
                    return;
                }

                // Convert the rotation to radians
                float angle = angleFS + (systemTime * fsAutoRot);
                angle = radians(angle);
                // Create a 2d rotation matrix
                mat2 rot = mat2(
                     cos(angle), sin(angle),
                    -sin(angle), cos(angle)
                );

                //-------------------------------------------------------------------
                // EMIT 4 vertices in order to display the Gfx element
                //-------------------------------------------------------------------

                // Type of GFX
                gfxType = fsType;
        
                // Emit a triangle strip creating a quad (4 vertices).
                // Here we need to make sure the rotation is applied before we position the sprite.
                // We just use hardcoded texture coordinates here. If an atlas is used we
                // can pass an additional vec4 for specific texture coordinates.
                // Each EmitVertex() emits values down the shader pipeline just like a single
                // run of a vertex shader, but in geomtry shaders we can do it multiple times!
                vec2 corner = vec2(0.0);
                // Anchor position is also affected by the scale value
                fsAnchor *= scaleFS;

                // Upper left
                corner = vec2(-hsize.x, -hsize.y) - fsAnchor;
                corner = rot * corner;
                corner = corner + center;
                gl_Position = projection * vec4(corner, 0.0, 1.0);
                uv = vec2(pos0.x, pos0.y);
                EmitVertex();

                // lower left
                corner = vec2(-hsize.x, hsize.y) - fsAnchor;
                corner = rot * corner;
                corner = corner + center;
                gl_Position = projection * vec4(corner, 0.0, 1.0);
                uv = vec2(pos0.x, pos1.y);
                EmitVertex();

                // upper right
                corner = vec2(hsize.x, -hsize.y) - fsAnchor;
                corner = rot * corner;
                corner = corner + center;
                gl_Position = projection * vec4(corner, 0.0, 1.0);
                uv = vec2(pos1.x, pos0.y);
                EmitVertex();

                // lower right
                corner = vec2(hsize.x, hsize.y) - fsAnchor;
                corner = rot * corner;
                corner = corner + center;
                gl_Position = projection * vec4(corner, 0.0, 1.0);
                uv = vec2(pos1.x, pos1.y);
                EmitVertex();

                // We are done with this triangle strip now
                EndPrimitive();
            }
        """

    def getFragment(self):
        return self._getHeader() + """
            uniform sampler2D  atlasTextureChan;

            in  vec2  uv;
            in  vec4  inColor;
            in  vec4  filterColor;
            in  float gfxType;
            out vec4  fragColor;

            vec3 RGB2YUV( vec3 rgb ){
                mat3 matrix = mat3(  0.299, -0.14713,  0.615  ,
                                     0.587, -0.28886, -0.51498,
                                     0.114,  0.436  , -0.10001 );
                return matrix * rgb;             
            }

            vec3 YUV2RGB( vec3 yuv ){
                mat3 matrix = mat3(  1.0   ,  1.0   , 1.0  ,
                                     0.0   , -0.3455, 1.779,
                                     1.4075, -0.7169, 0.0    );
                return matrix * yuv;             
            }

            void main() {

                // Get pixel color from the texture (only needed for Sprites
                vec4 color = texture(atlasTextureChan, uv);
                
                // Update color if basic shape
                if (gfxType == TYPE_RECTANGLE){
                    color = inColor;
                }
                
                // Modify color according to filter                
                vec3 pixel  = RGB2YUV(color.xyz);                
                vec3 fltClr = RGB2YUV(filterColor.xyz);

                // Lightness
                pixel.x = (pixel.x * fltClr.x);                
                // Color
                pixel.yz = (pixel.yz + fltClr.yz) * 0.5;

                // Transform to RGB back
                pixel = YUV2RGB(pixel);
                
                color.xyz = pixel;
                
                // Handle Transparency
                color.a  *= filterColor.a;

                // Set pixel color
                fragColor = color; 
            }
            """





class UniqueID():

    _uniqueID = 0
    @staticmethod
    def getID():
        res = UniqueID._uniqueID
        UniqueID._uniqueID += 1
        return res


class Box():

    def __init__(self, imgRsc, newId, index=0):
        # Prepare sub coords
        w = imgRsc.spriteW
        h = imgRsc.spriteH
        ix = index % imgRsc.nbSpriteX
        iy = index // imgRsc.nbSpriteX
        x0 = imgRsc.x0 + ix * w
        y0 = imgRsc.y0 + iy * h
        x1 = x0 + w
        y1 = y0 + h
        # Store box and image
        self._id = newId
        self._x = x0
        self._y = y0
        self._name = f"{imgRsc.name}"
        if imgRsc.nbSpriteX > 1 or imgRsc.nbSpriteY > 1:
            self._name += f"_{index}"
        self._img = imgRsc.data.crop((x0, y0, x1, y1))

    @property
    def x0(self):
        return self._x

    @property
    def y0(self):
        return self._y

    @x0.setter
    def x0(self, v):
        self._x = v

    @y0.setter
    def y0(self, v):
        self._y = v

    @property
    def width(self):
        return self._img.size[0]

    @property
    def height(self):
        return self._img.size[1]

    @property
    def x1(self):
        return self.x0 + self.width - 1

    @property
    def y1(self):
        return self.y0 + self.height - 1

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def data(self):
        return self._img

    def isOverlapping(self, other):
        return self.x0 <= other.x1 and self.x1 >= other.x0 and self.y0 <= other.y1 and self.y1 >= other.y0

    def __str__(self):
        return f"<Box id:{self.id} name:'{self.name}' - x:{self.x0} - y:{self.y0} - w:{self.width} - h:{self.height}>"


class Surface():

    def __init__(self, x0, y0, w, h):
        self._x0 = x0
        self._y0 = y0
        self._w  = w
        self._h  = h
        self._children = []

    @property
    def width(self):
        return self._w
    @property
    def height(self):
        return self._h
    @property
    def x0(self):
        return self._x0
    @property
    def y0(self):
        return self._y0
    @property
    def x1(self):
        return self.x0 + self.width - 1
    @property
    def y1(self):
        return self.y0 + self.height - 1
    @property
    def hasChildren(self):
        return len(self._children) > 0

    def __str__(self):
        return f"<SURFACE : {self.x0}/{self.y0}/{self.width}/{self.height}>"

    def increaseSize(self, oldMaxSize, newMaxSize):
        # Get X1 and Y1 limits
        oldX = oldMaxSize[0] - 1
        oldY = oldMaxSize[1] - 1
        newX = newMaxSize[0] - 1
        newY = newMaxSize[1] - 1
        # check if we have to stretch this surface
        if self.x1 == oldX:
            self._w += newX - oldX
        if self.y1 == oldY:
            self._h += newX - oldX
        # Do the same for child areas
        if self.hasChildren:
            self._children[0].increaseSize(oldMaxSize, newMaxSize)
            self._children[1].increaseSize(oldMaxSize, newMaxSize)

    def insertBox(self, box, border=1):
        # Get full box area
        boxW = box.width + 2 * border
        boxH = box.height + 2 * border
        # check if box fits into surface
        if self.width >= boxW and self.height >= boxH and not self.hasChildren:
            # Split surface and add children
            sB = Surface(self.x0, self.y0 + boxH, boxW, self.height - boxH)
            sR = Surface(self.x0 + boxW, self.y0, self.width - boxW, self.height)
            self._children.append(sR)
            self._children.append(sB)
            # Set position for the current box
            box.x0 = self.x0 + border
            box.y0 = self.y0 + border
            # Return OK
            return True
        elif self.hasChildren :
            # Check first child
            res1 = False
            res2 = False
            res1 = self._children[1].insertBox(box, border)
            if not res1:
                res2 = self._children[0].insertBox(box, border)
            return res1 or res2
        else:
            # no available surface
            return False



class BinPacking():

    # -----------------------------------------------------------------
    # Create box information from image list
    # -----------------------------------------------------------------
    def _createBoxes(self, imgList):
        #self._processInternal( (0,0,self._size-1, self._size-1) )
        boxes = []
        while len(imgList) > 0:
            # Get next image from list
            img     = imgList[0]
            imgList = imgList[1:]
            if img.nbSpriteX==1 and img.nbSpriteY==1:
                box = Box(img, 0)
                boxes.append(box)
            else:
                # Create new images and add them into the atlas
                for y in range(img.nbSpriteY):
                    for x in range(img.nbSpriteX):
                        idx = y*img.nbSpriteX + x
                        box = Box(img, UniqueID.getID(), idx)
                        boxes.append(box)
        return boxes

    # -----------------------------------------------------------------
    # Constructor
    # -----------------------------------------------------------------
    def __init__(self, squareSize, maxSize, imgBorder, resImgList):
        self._size   = squareSize
        self._border = imgBorder
        self._boxes  = self._createBoxes(resImgList)
        self._out = []
        self._surface = Surface(0, 0, self._size, self._size)
        self._maxSize = maxSize

    def _processInternal3(self):
        while len(self._boxes) > 0:
            # take first box
            box = self._boxes[0]
            # Split surfaces with current box
            isOK = self._surface.insertBox(box, self._border)
            if isOK:
                # Add new box in output
                self._out.append(box)
                # Remove box from TODO list
                self._boxes = self._boxes[1:]
            else:
#                print( f"Impossible to add box {box}")
#                print( "Increase squareSize")
#                print(f"old surface {self._surface.width}-{self._surface.height}")
                addS = max(box.width, box.height)
                oldS = (self._surface.width, self._surface.height)
                newS = (int(self._surface.width  + addS + 2*self._border),
                        int(self._surface.height + addS + 2*self._border))
                self._surface.increaseSize(oldS, newS)
#                print(f"new surface {self._surface.width}-{self._surface.height}")
                if self._surface.width > self._maxSize:
                    return False
        return True


    def process(self):
        self._out = []
        res = self._processInternal3()
        return res, self._out, max(self._surface.width, self._surface.height)







class Shader():

    def __init__(self, name="NO_SHADER"):
        self._name = name

    @property
    def name(self):
        return self._name

    def getVertex(self):
        raise RuntimeError("Shader::getVertex() has not been implemented yet")
    def getGeometry(self):
        raise RuntimeError("Shader::getGeometry() has not been implemented yet")
    def getFragment(self):
        raise RuntimeError("Shader::getFragment() has not been implemented yet")







class Component():

    TYPE_GFX          = 1
    TYPE_SCRIPT       = 2
    TYPE_KEY          = 3
    TYPE_PAD_BUTTON   = 4
    TYPE_MOUSE_BUTTON = 5
    TYPE_PAD_AXIS     = 6


    # ------------------------------------
    #  SLOTS
    # ------------------------------------
    __slots__ = ['_name',
                 '_type',
                 '_entity'
                ]

    def __init__(self, type, name="Component"):
        self._entity = None
        self._name   = name
        self._type   = type

    def getName(self):
        return self._name

    def getType(self):
        return self._type

    def getPriority(self):
        return 0

    # ------------------------------------
    #  ENTITY LINK
    # ------------------------------------
    def setEntity(self, ent):
        if self._entity != None:
            raise RuntimeError(f"[ERROR] Impossible to make a link between this component {self} and another entity !")
        self._entity = ent

    def resetEntity(self, ent):
        if self._entity != ent:
            raise RuntimeError(f"[ERROR] Impossible to remove a link between this component {self} and its entity !")
        self._entity = None

    # ------------------------------------
    #  SCENE LINK
    # ------------------------------------
    def getScene(self):
        res = None
        if self._entity != None:
            res = self._entity.getScene()
        return res



import math
from array import array

import moderngl
import moderngl_window
from pyrr import Matrix44

from ..components.gfx import Gfx
from ..gpu.fsgpu_main import FsGpuMain
from ..shaders.simple_shader import SimpleShader
from ..systems.gfx_system import GfxSystem


class OpenGLData():

    # ========================================================
    #  PARAMETERS
    # ========================================================
    DEBUG_DISPLAY_QUERY   = False
    DEBUG_DISPLAY_FSGPU   = False
    CHANNEL_ATLAS_INFO    = 0
    CHANNEL_ATLAS_TEXTURE = 1
    CHANNEL_FILE_SYSTEM   = 2


    # ========================================================================
    #  PRIVATE METHODS
    # ========================================================================
    def _initCfg(self):
        # GL context
        self.ctx = moderngl.create_context(require=330)
        moderngl_window.activate_context(ctx=self.ctx)
        # the selected program will be used in the prepareData method
        self._program = None
        self._programs = {}
        # elapsed time
        self._elapsedTime = 0
        # TODO : remove this line when moderngl is 5.7.0
        self.ctx.info["GL_MAX_GEOMETRY_OUTPUT_VERTICES"] = 256
        # Display minimal information on GL Device
        print("\nYou are using the following GL device :")
        for field in ["GL_VENDOR",
                      "GL_RENDERER",
                      "GL_VERSION",
                      "GL_MAX_TEXTURE_SIZE",
                      "GL_MAX_GEOMETRY_OUTPUT_VERTICES",
                      ]:
            hardCoded = ""
            if field == "GL_MAX_GEOMETRY_OUTPUT_VERTICES":
                hardCoded = "(/!\ HARD-CODED /!\)"
            print(f"    - {field} : {self.ctx.info[field]} {hardCoded}")
        print()

    def _initFsGPU(self, fsGpuMemSize):

        # -----------------------------------------------------------------
        # SHADERS
        # -----------------------------------------------------------------
        # Instanciate shader object and add program
        shader = SimpleShader()
        self.addProgram(
                "Program #1",
                vertexStr=shader.getVertex(),
                geometryStr=shader.getGeometry(),
                fragmentStr=shader.getFragment()  )

        # -----------------------------------------------------------------
        # GPU FILE SYSTEM
        # -----------------------------------------------------------------
        # the height indicates number of pages
        # the width the size of each 1-height FSGpuBuffer
        # Using the start/end allocation mode, the static sprites could be
        # stored in the last pages, while the dynamic ones could be stored
        # in the first pages. So that would lead to save some page update
        # adn though less writing operations
        # -----------------------------------------------------------------
        # Data area = Width * Height (will be multiplied by number of components (four 32-bit float values))
        # That means sizeW and sizeH represent the number of pixel of the texture.
        # for one pixel, there are values, related to RGBA layers, and each value
        # is a 32-bit float
        sizeW = min(16 * 1024, self.ctx.info["GL_MAX_TEXTURE_SIZE"])
        sizeH = fsGpuMemSize // (sizeW * 4)
        # instanciate FsGpu
        print(f"\nCreating FsGPU texture : size={sizeW} x {sizeH} x 4 components (32-bit float values)")
        self._fsgpu = FsGpuMain(self.ctx, sizeW, sizeH)

        # -----------------------------------------------------------------
        # SPRITE BUFFER (VERTEX)
        # -----------------------------------------------------------------
        # Create sprite information (input data)
        nbComponents = 4
        spriteInfoSize = (1) * nbComponents  # 1 value in vertex input per sprite (the File System id)
        # Maximum set to 1 million gfx elements
        nbMaxSprites   = 1000000
        self._vertexBuffer = self.ctx.buffer(reserve=nbMaxSprites * spriteInfoSize)
        # Number of useful data in the vertex buffer
        self._nbSprites = 0

        # -----------------------------------------------------------------
        # VIEWPORT
        # -----------------------------------------------------------------
        # Change view port (for static sprites)
        w, h = self.ctx.screen.size
        self.setViewPort(0, 0, w, h)

        # -----------------------------------------------------------------
        # VERTEX ARRAY OBJECT
        # -----------------------------------------------------------------
        # Create vertex array object
        self._vao = self.ctx.vertex_array(
                        self._program,
                        [
                            (self._vertexBuffer,
                             "1u",
                             "blockID"),
                        ]
                    )

        # -----------------------------------------------------------------
        # UNIFORMS DATA
        # -----------------------------------------------------------------
        # Uniform data to know from which channel the data can be read
        # - texture atlas Information (where are located the textures)
        # - texture atlas
        # - gpu file system
        # - system time (start at 0)
        self._program["atlasTextureChan"] = OpenGLData.CHANNEL_ATLAS_TEXTURE
        self._program["atlasInfoChan"   ] = OpenGLData.CHANNEL_ATLAS_INFO
        self._program["fsGpuChan"       ] = OpenGLData.CHANNEL_FILE_SYSTEM
        self._program["systemTime"      ] = 0

        # -----------------------------------------------------------------
        # PROFILING
        # -----------------------------------------------------------------
        # Query configuration for profiling
        if OpenGLData.DEBUG_DISPLAY_QUERY:
            self._query = self.ctx.query(samples=True, time=True, primitives=True)

        # -----------------------------------------------------------------
        # GFX
        # -----------------------------------------------------------------
        Gfx.setLoader(self._loader)
        Gfx.setFsGPU(self._fsgpu)
        GfxSystem.setOpenGlData(self)

    def _loadAtlas(self):
        # -----------------------------------------------------------------
        # IMAGE ATLAS
        # -----------------------------------------------------------------
        # compute image atlas from the resource loader
        maxSide  = self.ctx.info["GL_MAX_TEXTURE_SIZE"]
        sizeSide = min(128, maxSide)
        self._loader.generateImageAtlas(sizeSide, maxSide, border=1)
        # Load fonts into the GPU File System
        self._loader.storeFonts(self._fsgpu)
        # Create atlas texture
        image = self._loader.getTextureImage()
        self._atlasImages = self.ctx.texture(image.size, 4, image.tobytes())

        # -----------------------------------------------------------------
        # ATLAS INFO
        # -----------------------------------------------------------------
        # Set texture to contain all texture info (needed to access atlas data)
        # Write Pyrogen Texture data (20 x 32bit values)
        # Pack values using the nb of components
        # e.g.  X,Y,W,H packed in a single RGBA value
        # texelFetch() function to get them from GPU side
        # texture.filter = moderngl.NEAREST, moderngl.NEAREST
        #
        # [FOR EACH TEXTURE]
        # > Diffuse data
        #    - (X,Y) top left position
        #    - (W,H) size
        nbTextures       = self._loader.getNbTextures()
        nbDataPerTexture = 1
        nbComponents     = 4
        width            = nbTextures * nbDataPerTexture
        nbBytes          = width * nbComponents
        # Prepare texture
        self._atlasInfo = self.ctx.texture((width, 1), nbComponents, dtype="u4")
        self._atlasInfo.filter = (moderngl.NEAREST, moderngl.NEAREST)

        # Write texture positions and sizes
        data = array("I", [0,] * (nbBytes) )
        allIDs = self._loader.getAllIds()
        for id in allIDs:
            data[4*id+0] = self._loader.getTextureById(id)["x"]
            data[4*id+1] = self._loader.getTextureById(id)["y"]
            data[4*id+2] = self._loader.getTextureById(id)["w"]
            data[4*id+3] = self._loader.getTextureById(id)["h"]
        self._atlasInfo.write(data.tobytes())


    # ========================================================================
    #  CONSTRUCTOR
    # ========================================================================
    def __init__(self, loader, fsGpuMemSize):
        # Create window and context
        self._initCfg()
        # Store resource loader reference
        self._loader = loader
        # Init GPU
        self._initFsGPU(fsGpuMemSize)

    def createAtlas(self):
        self._loadAtlas()

    def update(self, deltaTime, systemTime):
        # update elapsed time and copy it into the uniform var
        self._elapsedTime = systemTime
        self._program["systemTime"] = self._elapsedTime

        # Process File system
        self._fsgpu.update(deltaTime)

    def render(self):
        # update file system texture
        self._fsgpu.render()

        # Clear buffer with background color
        self.ctx.clear(
            (math.sin(self._elapsedTime + 0) + 1.0) / 4 + 0.5,
            (math.sin(self._elapsedTime + 2) + 1.0) / 4 + 0.5,
            (math.sin(self._elapsedTime + 3) + 1.0) / 4 + 0.5,
        )

        # Set blitting mode
        # self.ctx.blend_equation  = moderngl.FUNC_SUBTRACT
        # self.ctx.blend_func = (moderngl.ONE, moderngl.ONE)
        self.ctx.enable(moderngl.BLEND)

        # Set the atlas texture to an opengl channel
        self._atlasInfo.use(OpenGLData.CHANNEL_ATLAS_INFO)
        self._atlasImages.use(OpenGLData.CHANNEL_ATLAS_TEXTURE)
        self._fsgpu.getTexture().use(OpenGLData.CHANNEL_FILE_SYSTEM)

        # Process rendering
        if not OpenGLData.DEBUG_DISPLAY_QUERY:
            self._vao.render(mode=moderngl.POINTS, vertices=self._nbSprites)
        else:
            with self._query:
                self._vao.render(mode=moderngl.POINTS, vertices=self._nbSprites)
                if self.ctx.error != "GL_NO_ERROR":
                    print("[ERROR] during rendering...")
                    print(self.ctx.error)
                    exit()


    # ========================================================================
    #  VIEWPORT CONFIGURATION
    # ========================================================================
    def setViewPort(self, x0, y0, x1, y1):
        near = 1
        far  = 0
        matrix = Matrix44.orthogonal_projection(x0, x1, y1, y0, near, far, dtype="f4")
        self._program["projection"].write(matrix)


    # ========================================================================
    #  VIEWPORT CONFIGURATION
    # ========================================================================
    def updateVertexBuffer(self, data, nbSprites):
        self._vertexBuffer.write(data)
        self._nbSprites = nbSprites


    # ========================================================================
    #  SHADER PROGRAMS
    # ========================================================================
    # The first program to be added, is automatically selected
    def addProgram(self, name, vertexStr="", geometryStr="", fragmentStr=""):
        # Create program from shader sources
        prog = self.ctx.program(vertex_shader=vertexStr,
                                geometry_shader=geometryStr,
                                fragment_shader=fragmentStr)
        # check this program name is not already used
        if name in self._programs:
            raise RuntimeError("[ERROR] loadProgram : {name} is already in the program list !")
        # Store new program
        self._programs[name] = prog
        # Select this program is this is the first one
        if len(self._programs) == 1:
            self.selectProgram(name)

    # Select a program from the program list
    def selectProgram(self, name):
        if name in self._programs:
            self._program = self._programs[name]


    # ========================================================================
    #  DEBUG
    # ========================================================================
    def displayFS(self):
        if OpenGLData.DEBUG_DISPLAY_FSGPU:
            self._fsgpu.display()


from .component import Component

class Script(Component):

    __slots__ = ["_priority",]

    def __init__(self, priority=0, name="Script"):
        super().__init__(Component.TYPE_SCRIPT, name)
        # Store priority value
        self._priority = priority

    def updateScript(self, deltaTime, systemTime):
        raise RuntimeError("[ERROR] updateScript() method has not been implemented yet !")

    def setPriority(self, newP):
        self._priority = newP
        scn = self.getScene()
        if scn != None:
            scn.notifyChangeScriptPriority(self)

    def getPriority(self):
        return self._priority




from .world import World


class Scene():

    # ========================================================
    # Constructor
    # ========================================================
    def __init__(self):
        # Create a world entity
        self._world = World()
        # Init scene manager ref
        self._scnMgr = None
        # Entity list
        self._entitiesByName = {}
        self._entitiesByRef  = []
        # System time for this scene
        self._systemTime = 0

    # ========================================================
    # Application process
    # ========================================================
    def setManager(self, ref):
        self._scnMgr = ref

    def updateWorld(self, deltaTime):
        self._systemTime += deltaTime
        self._world.updateSystems(deltaTime, self._systemTime)

    def renderWorld(self):
        self._world.renderSystems()

    # ========================================================
    # Scene management
    # ========================================================
    def registerScene(self, name, ref):
        self._scnMgr.registerScene(name, ref)

    def unregisterScene(self, name):
        self._scnMgr.unregisterScene(name)

    def selectScene(self, name):
        self._scnMgr.selectScene(name)

    # ========================================================
    # Entity management
    # ========================================================
    def addEntity(self, ref):
        # Check if we can add the entity
        if ref in self._entitiesByRef:
            raise RuntimeError(f"[ERROR] Cannot add entity {ref} into the scene {self} !")
        # Add the entity
        self._entitiesByRef.append(ref)
        self._entitiesByName[ref.getName()] = ref
        # Create the link between scene and entity
        ref.setScene(self)

    def removeEntity(self, ref):
        # Check if we can add the entity
        if ref not in self._entitiesByRef:
            raise RuntimeError(f"[ERROR] Cannot remove entity {ref} from the scene {self} !")
        # Remove the link between the entoty and the current scene
        ref.resetScene(self)
        # Remove the entity
        self._entitiesByRef.remove(ref)
        del self._entitiesByName[ref.getName()]

    def getNbEntities(self):
        return len(self._entitiesByRef)

    # ========================================================
    # Component management
    # ========================================================
    def registerComponent(self, ref):
        self._world.registerComponent(ref)

    def unregisterComponent(self, ref):
        self._world.unregisterComponent(ref)


    # ========================================================
    # Specific events
    # ========================================================
    def notifyChangeZ(self, ref):
        self._world.notifyChangeZ(ref)

    def notifyChangeScriptPriority(self, ref):
        self._world.notifyChangeScriptPriority(ref)


    # ========================================================
    # Events
    # ========================================================
    def keyboardEvent(self, keyID, isPressed, modifiers):
        self._world.keyboardEvent(keyID, isPressed, modifiers)

    def mouseButtonEvent(self, x, y, buttonID, isPressed, modifiers):
        self._world.mouseButtonEvent(x, y, buttonID, isPressed, modifiers)

    def mouseMotionEvent(self, x, y, dx, dy):
        self._world.mouseMotionEvent(x, y, dx, dy)

    def mouseDragEvent(self, x, y, dx, dy, buttonID, modifiers):
        self._world.mouseDragEvent(x, y, dx, dy, buttonID, modifiers)

    def mouseScrollEvent(self, x, y, dx, dy):
        self._world.mouseScrollEvent(x, y, dx, dy)

    def gamepadButtonEvent(self, gamepadID, buttonID, isPressed):
        self._world.gamepadButtonEvent(gamepadID, buttonID, isPressed)

    def gamepadAxisEvent(self, gamepadID, axisID, analogValue):
        self._world.gamepadAxisEvent(gamepadID, axisID, analogValue)



from pyrogen.src import pyrogen
import setuptools


setuptools.setup(version=pyrogen.__version__,
                 include_package_data=True,
                 install_requires=[
                     "pyglet==1.5.17",
                     "pillow==8.1.2",
                     "moderngl==5.6.4",
                     "moderngl-window==2.3.0", ]
                 )



from ..components.component  import Component
from ..systems.gfx_system    import GfxSystem
from ..systems.script_system import ScriptSystem
from ..systems.input_system import KeyboardSystem, GamepadButtonSystem, GamepadAxisSystem, MouseSystem


class World():

    # Constructor
    def __init__(self):
        # GFX
        self._gfxSys     = GfxSystem()
        # SCRIPT
        self._scrSys     = ScriptSystem()
        # INPUT
        self._keySys     = KeyboardSystem()
        self._padButtSys = GamepadButtonSystem()
        self._padAxisSys = GamepadAxisSystem()
        self._mouseSys   = MouseSystem()


    # Application process
    def updateSystems(self, deltaTime, systemTime):
        self._gfxSys.update(deltaTime, systemTime)
        self._scrSys.updateScripts(deltaTime, systemTime)

    def renderSystems(self):
        self._gfxSys.render()


    # Component registering
    def registerComponent(self, ref):
        # Check component type
        type = ref.getType()
        if type == Component.TYPE_GFX:
            self._gfxSys.addComponent(ref)
        elif type == Component.TYPE_SCRIPT:
            self._scrSys.addComponent(ref)
        elif type== Component.TYPE_KEY:
            self._keySys.addComponent(ref)
        elif type == Component.TYPE_PAD_BUTTON:
            self._padButtSys.addComponent(ref)
        elif type == Component.TYPE_PAD_AXIS:
            self._padAxisSys.addComponent(ref)
        elif type == Component.TYPE_MOUSE_BUTTON:
            self._mouseSys.addComponent(ref)
        else:
            raise RuntimeError("[ERROR] Cannot add the component in any system !")

    def unregisterComponent(self, ref):
        print(f"Unregistering a component : {ref} is not IMPLEMENTED YET !!!")
        pass


    # Specific events
    def notifyChangeZ(self, ref):
        self._gfxSys.notifyChangeZ(ref)

    def notifyChangeScriptPriority(self, ref):
        self._scrSys.notifyChangeScriptPriority(ref)


    # Input Events
    def keyboardEvent(self, keyID, isPressed, modifiers):
        self._keySys.keyboardEvent(keyID, isPressed, modifiers)

    def mouseButtonEvent(self, x, y, buttonID, isPressed, modifiers):
        self._mouseSys.mouseButtonEvent(x, y, buttonID, isPressed, modifiers)

    def mouseMotionEvent(self, x, y, dx, dy):
        # TODO
        pass

    def mouseDragEvent(self, x, y, dx, dy, buttonID, modifiers):
        # TODO
        pass

    def mouseScrollEvent(self, x, y, dx, dy):
        # TODO
        pass

    def gamepadButtonEvent(self, gamepadID, buttonID, isPressed):
        self._padButtSys.gamepadButtonEvent(gamepadID, buttonID, isPressed)

    def gamepadAxisEvent(self, gamepadID, axisID, analogValue):
        self._padAxisSys.gamepadAxisEvent(gamepadID, axisID, analogValue)




from ....ecs3.main.scene       import Scene
from ....ecs3.main.entity      import Entity
from ....ecs3.main.common      import Buttons, Gamepads, Axis
from ....ecs3.components.gfx   import GfxSprite, GfxBox
from ....ecs3.components.input import Keyboard, GamepadButton, MouseButton, GamepadAxis

from ..components.scripts import MoveCircle, MoveSquare, ShowHide, MoveShip

import pyglet.window.key as keys
import random


class FpsTest(Scene):

    def __init__(self):
        super().__init__()

        # ----------------------------------------------
        # Ship entity
        # ----------------------------------------------
        ship = Entity()
        # Create components
        shipGfx = GfxSprite(f"ship001")
        shipGfx.setX(480)
        shipGfx.setY(540)
        shipGfx.setZIndex(20000)
        shipGfx.setScale(1.0)
        axis = GamepadAxis(0.1)
        axis.addAxis(Gamepads.ANY, Axis.LX, "moveX")
        axis.addAxis(Gamepads.ANY, Axis.LY, "moveY")
        axis.addAxis(Gamepads.ANY, Axis.RX, "moveX")
        axis.addAxis(Gamepads.ANY, Axis.RY, "moveY")
        moveShip = MoveShip(shipGfx, axis)
        # add components to entity
        ship.addComponent(shipGfx)
        ship.addComponent(axis)
        ship.addComponent(moveShip)
        # Add ship to scene
        self.addEntity(ship)

        # ----------------------------------------------
        # Main Sprite
        # ----------------------------------------------
        mainCharacter = Entity()
        # Create components
        mySprite = GfxSprite(f"characters_0")
        mySprite.setX(960)
        mySprite.setY(540)
        mySprite.setZIndex(10000)
        mySprite.setScale(4)
        padButtons = GamepadButton()
        padButtons.addButton(Gamepads.ANY, Buttons.A  , "Hide"   )
        padButtons.addButton(Gamepads.ANY, Buttons.B  , "Show"   )
        padButtons.addButton(Gamepads.ANY, Buttons.LB , "Enlarge")
        padButtons.addButton(Gamepads.ANY, Buttons.RB , "Reduce" )
        mouseButtons = MouseButton()
        mouseButtons.addButton(Buttons.MOUSE_LEFT , "RotateLeft" )
        mouseButtons.addButton(Buttons.MOUSE_RIGHT, "RotateRight")
        scrShowHide = ShowHide(mySprite, padButtons, mouseButtons)
        scrShowHide.setPriority(10)
        # add components to entity
        mainCharacter.addComponent(mySprite)
        mainCharacter.addComponent(padButtons)
        mainCharacter.addComponent(mouseButtons)
        mainCharacter.addComponent(scrShowHide)
        # add entity to scene
        self.addEntity(mainCharacter)

        # ----------------------------------------------
        # Other Gfx components
        # ----------------------------------------------
        NX = 70
        NY = 35
        N = NX * NY

        for i in range(N):

            # -------------------------------------
            # Create Entity
            entity = Entity()

            # -------------------------------------
            # Create Sprite or plain color rectangle
            clr = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
            if i%2 == 0:
                index = random.randint(0,95)
                mySprite = GfxSprite(f"characters_{index}")
                mySprite.setColor(clr)
            else:
                w = random.randint(30,34)
                h = random.randint(46,50)
                mySprite = GfxBox(inClr=clr, width=w, height=h)
            # Position
            x = (i %  NX) * (1920//NX)
            y = (i // NX) * (1080//NY)
            mySprite.setX(x)
            mySprite.setY(y)
            mySprite.setZIndex(i+100)
#            mySprite.setAnchor(-16,-24)
            mySprite.setAutoRotate( random.randint(0,360) )
#            mySprite.setAngle( random.randint(0,360) )
#            mySprite.setFlipY(True)
            tALL = random.random() * 2 + 1.0
            tON  = random.random() * (tALL-0.1) + 0.1
            mySprite.setVisibility( tON, tALL )
#            mySprite.setScale(1.5)

            # -------------------------------------
            # Create script to move sprite
            radius = random.randint(30,300)
            speed  = random.randint(-360,360)

            # -------------------------------------
            # Add input components to move all the sprites
            keyboard = Keyboard()
            if i%2 == 1:
                keyboard.addKey(keys.UP, "moveUp")
                keyboard.addKey(keys.DOWN, "moveDown")
                scrMove = MoveCircle(mySprite, (x,y), radius, speed, keyboard)
            else:
                keyboard.addKey(keys.LEFT, "moveLeft")
                keyboard.addKey(keys.RIGHT, "moveRight")
                scrMove = MoveSquare(mySprite, (x, y), radius, speed, keyboard)

            # -------------------------------------
            # Add all components into entity
            entity.addComponent(mySprite)
            entity.addComponent(keyboard)
            entity.addComponent(scrMove)

            # -------------------------------------
            # Add entity to the Scene (containing both components)
            self.addEntity(entity)




# __init__.py (pyrogen package)
from .version import __version__

# Main
from .ecs3.main.pyrogen_app import Pyrogen
from .ecs3.main.entity      import Entity
from .ecs3.main.scene       import Scene

# Components
from .ecs3.components.gfx    import GfxSprite, GfxBox
from .ecs3.components.script import Script
from .ecs3.components.input  import Keyboard, GamepadButton, MouseButton


